import importlib.resources
import json
import logging
import os
import platform
import re
import subprocess
import tempfile
import time
from typing import List, Optional

import boto3
import typer
import yaml

from autogluon.bench import __path__ as agbench_path
from autogluon.bench import __version__ as agbench_version
from autogluon.bench.cloud.aws.stack_handler import deploy_stack, destroy_stack
from autogluon.bench.eval.hardware_metrics.hardware_metrics import get_hardware_metrics
from autogluon.bench.frameworks.multimodal.multimodal_benchmark import MultiModalBenchmark
from autogluon.bench.frameworks.tabular.tabular_benchmark import TabularBenchmark
from autogluon.bench.frameworks.timeseries.timeseries_benchmark import TimeSeriesBenchmark
from autogluon.bench.utils.general_utils import (
    download_dir_from_s3,
    download_file_from_s3,
    formatted_time,
    upload_to_s3,
)

app = typer.Typer()

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
AMLB_DEPENDENT_MODULES = ["tabular", "timeseries"]
INDEPENDENT_MODULES = ["multimodal"]

with importlib.resources.path("autogluon.bench", "Dockerfile") as docker_file:
    project_path = os.path.join(os.path.dirname(docker_file))


def get_kwargs(module: str, configs: dict):
    """Returns a dictionary of keyword arguments to be used for setting up and running the benchmark.

    Args:
        module (str): The name of the module to benchmark (either "multimodal", "tabular", or "timeseires").
        configs (dict): A dictionary of configuration options for the benchmark.

    Returns:
        A dictionary containing the keyword arguments to be used for setting up and running the benchmark.
    """

    if module in INDEPENDENT_MODULES:
        framework_configs = get_framework_configs(configs=configs)
        return {
            "setup_kwargs": {
                "git_uri": framework_configs["repo"],
                "git_branch": framework_configs.get("version", "stable"),
            },
            "run_kwargs": {
                "dataset_name": configs["dataset_name"],
                "framework": configs["framework"],
                "constraint": configs.get("constraint"),
                "params": framework_configs.get("params"),
                "custom_dataloader": configs.get("custom_dataloader"),
                "custom_metrics": configs.get("custom_metrics"),
                "time_limit": configs.get("time_limit"),
            },
        }
    elif module in AMLB_DEPENDENT_MODULES:
        git_uri, git_branch = _get_git_info(configs["git_uri#branch"])
        return {
            "setup_kwargs": {
                "git_uri": git_uri,
                "git_branch": git_branch,
                "framework": configs["framework"],
                "user_dir": configs.get("amlb_user_dir"),
            },
            "run_kwargs": {
                "framework": configs["framework"],
                "benchmark": configs.get("amlb_benchmark", "test"),
                "constraint": configs.get("amlb_constraint", "test"),
                "task": configs.get("amlb_task"),
                "fold": configs.get("fold_to_run"),
                "user_dir": configs.get("amlb_user_dir"),
            },
        }


def _get_benchmark_name(configs: dict) -> str:
    default_benchmark_name = "ag_bench"
    benchmark_name = configs.get("benchmark_name", default_benchmark_name) or default_benchmark_name
    return benchmark_name


def run_benchmark(
    benchmark_name: str, benchmark_dir: str, configs: dict, benchmark_dir_s3: str = None, skip_setup: str = False
):
    """Runs a benchmark based on the provided configuration options.

    Args:
        configs (dict): A dictionary of configuration options for the benchmark.

    Returns:
        None
    """

    module_to_benchmark = {
        "multimodal": MultiModalBenchmark,
        "tabular": TabularBenchmark,
        "timeseries": TimeSeriesBenchmark,
    }
    module_name = configs["module"]

    benchmark_class = module_to_benchmark.get(module_name, None)
    if benchmark_class is None:
        raise NotImplementedError

    benchmark = benchmark_class(benchmark_name=benchmark_name, benchmark_dir=benchmark_dir)

    module_kwargs = get_kwargs(module=module_name, configs=configs)
    if not skip_setup:
        benchmark.setup(**module_kwargs["setup_kwargs"])

    benchmark.run(**module_kwargs["run_kwargs"])
    logger.info(f"Backing up benchmarking configs to {benchmark.metrics_dir}/configs.yaml")
    _dump_configs(benchmark_dir=benchmark.metrics_dir, configs=configs, file_name="configs.yaml")

    if configs.get("METRICS_BUCKET", None):
        framework_name = module_kwargs["run_kwargs"]["framework"].split(":")[0]
        upload_path = []
        if module_name == "multimodal":
            upload_path = [
                configs.get("mode", None),
                module_kwargs["run_kwargs"].get("dataset_name", None),
                module_kwargs["run_kwargs"].get("constraint", None),
                framework_name,
            ]
        elif module_name == "tabular" or module_name == "timeseries":
            upload_path = [
                configs.get("mode", None),
                module_kwargs["run_kwargs"].get("benchmark", None),
                module_kwargs["run_kwargs"].get("constraint", None),
                module_kwargs["run_kwargs"].get("task", None),
                module_kwargs["run_kwargs"].get("fold", None),
                framework_name,
            ]
        else:  # ToDo: Re-order fields for specific module name if required
            upload_path = [
                configs.get("mode", None),
                module_kwargs["run_kwargs"].get("dataset_name", None),
                module_kwargs["run_kwargs"].get("constraint", None),
                framework_name,
            ]

        upload_path_str = ".".join(str(part) for part in upload_path)
        benchmark_dir_s3 += f"/{upload_path_str}"
        benchmark.upload_metrics(s3_bucket=configs["METRICS_BUCKET"], s3_dir=benchmark_dir_s3)


def invoke_lambda(configs: dict, config_file: str) -> dict:
    """Invokes an AWS Lambda function to run benchmarks based on the provided configuration options.

    Args:
        configs (dict): A dictionary of configuration options for the AWS infrastructure.
        config_file (str): The path of the configuration file to use for running the benchmarks.

    Returns:
        A dictionary containing Lambda response payload.
    """

    lambda_client = boto3.client("lambda", configs["CDK_DEPLOY_REGION"])
    payload = {"config_file": config_file}
    lambda_function_name = configs["LAMBDA_FUNCTION_NAME"] + "-" + configs["STACK_NAME_PREFIX"]
    response = lambda_client.invoke(
        FunctionName=lambda_function_name, InvocationType="RequestResponse", Payload=json.dumps(payload)
    )
    response_payload = json.loads(response["Payload"].read().decode("utf-8"))

    if "FunctionError" in response:
        error_payload = response_payload
        raise Exception(f"Lambda function error: {error_payload}")

    logger.info("AWS Batch jobs submitted by %s.", configs["LAMBDA_FUNCTION_NAME"])

    return response_payload


@app.command()
def get_job_status(
    job_ids: Optional[List[str]] = typer.Option(None, "--job-ids", help="List of job ids, separated by space."),
    cdk_deploy_region: Optional[str] = typer.Option(None, help="AWS region that the Batch jobs run in."),
    config_file: Optional[str] = typer.Option(None, help="Path to YAML config file containing job ids."),
):
    """
    Query the status of AWS Batch job ids.
    The job ids can either be passed in directly or read from a YAML configuration file.

    Args:
        job_ids (List[str], optional):
            A list of job ids to query the status for.
        cdk_deploy_region (str, optional):
            AWS region that the Batch jobs run in.
        config_file (str, optional):
            A path to a YAML config file containing job ids. The YAML file should have the structure:
                job_configs:
                    <job_id>: <job_config>
                    <job_id>: <job_config>
                    ...

    Returns:
        A dictionary containing the status of the queried job ids.
    """
    if config_file is not None:
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)
            job_ids = list(config.get("job_configs", {}).keys())
            cdk_deploy_region = config.get("CDK_DEPLOY_REGION", cdk_deploy_region)
    if job_ids is None or cdk_deploy_region is None:
        raise ValueError("Either job_ids or cdk_deploy_region must be provided or configured in the config_file.")

    batch_client = boto3.client("batch", region_name=cdk_deploy_region)

    status_dict = {}

    for job_id in job_ids:
        response = batch_client.describe_jobs(jobs=[job_id])
        job_detail = response["jobs"][0]

        # Check if the job is an array job
        if "arrayProperties" in job_detail and "size" in job_detail["arrayProperties"]:
            status_dict[job_id] = job_detail["arrayProperties"]["statusSummary"]
        else:
            status_dict[job_id] = job_detail["status"]

    logger.info(status_dict)
    return status_dict


def wait_for_jobs(
    config_file: Optional[str] = None,
    job_ids: Optional[List[str]] = None,
    aws_region: Optional[str] = None,
    quit_statuses: Optional[list] = ["SUCCEEDED", "FAILED"],
    frequency: Optional[int] = 120,
):
    if config_file is not None:
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)
            job_ids = list(config.get("job_configs", {}).keys())
            aws_region = config.get("CDK_DEPLOY_REGION", aws_region)

    batch_client = boto3.client("batch", region_name=aws_region)
    failed_jobs = set()

    while True:
        all_jobs_completed = True
        job_status = get_job_status(job_ids=job_ids, cdk_deploy_region=aws_region, config_file=None)

        for job_id, status in job_status.items():
            if isinstance(status, str):
                if status == "FAILED":
                    failed_jobs.add(job_id)
                elif status not in quit_statuses:
                    all_jobs_completed = False
            elif isinstance(status, dict):
                for status, num in status.items():
                    if status == "FAILED" and num > 0:
                        paginator = batch_client.get_paginator("list_jobs")
                        for page in paginator.paginate(arrayJobId=job_id, jobStatus="FAILED"):
                            for job in page["jobSummaryList"]:
                                failed_jobs.add(job["jobId"])
                    if status not in quit_statuses and num > 0:
                        all_jobs_completed = False

        if all_jobs_completed:
            break
        else:
            time.sleep(frequency)  # Poll job statuses every 120 seconds

    return failed_jobs


def _dump_configs(benchmark_dir: str, configs: dict, file_name: str):
    os.makedirs(benchmark_dir, exist_ok=True)
    config_path = os.path.join(benchmark_dir, file_name)
    with open(config_path, "w") as file:
        yaml.dump(configs, file)
        logger.info(f"Configs have been saved to {config_path}")
    return config_path


def _get_git_info(git_uri_branch: str):
    git_info = git_uri_branch.split("#")
    if len(git_info) == 2:
        git_uri, git_branch = git_info
    elif len(git_info) == 1:
        git_uri = git_info[0]
        git_branch = "stable"
    return git_uri, git_branch


def _is_mounted(path: str):
    with open("/proc/mounts", "r") as mounts:
        for line in mounts:
            parts = line.split()
            if len(parts) > 1 and parts[1] == path:
                return True


def _umount_if_needed(path: str = None):
    if not path:
        return

    if _is_mounted(path):
        logging.info(f"{path} is a mount point, attempting to umount.")
        subprocess.run(["sudo", "umount", path])
        logging.info(f"Successfully umounted {path}.")


def _mount_dir(orig_path: str, new_path: str):
    logging.info(f"Mounting from {orig_path} to {new_path}.")
    subprocess.run(["sudo", "mount", "--bind", orig_path, new_path])


def update_custom_dataloader(configs: dict):
    custom_dataloader_file = configs["custom_dataloader"]["dataloader_file"]
    original_path = os.path.abspath(os.path.dirname(custom_dataloader_file))
    custom_dataset_config = configs["custom_dataloader"]["dataset_config_file"]
    if original_path != os.path.abspath(os.path.dirname(custom_dataset_config)):
        raise ValueError(
            "Custom dataloader dataset definition <config_file> and dataloader definition <file_path> need to be placed under the same parent directory."
        )
    dataloader_file_name = os.path.basename(custom_dataloader_file)
    dataset_config_file_name = os.path.basename(custom_dataset_config)
    custom_dataloader_path = os.path.join(project_path, "custom_configs", "dataloaders")

    configs["custom_dataloader"]["dataloader_file"] = f"custom_configs/dataloaders/{dataloader_file_name}"
    configs["custom_dataloader"]["dataset_config_file"] = f"custom_configs/dataloaders/{dataset_config_file_name}"

    return original_path, custom_dataloader_path


def update_custom_metrics(configs: dict):
    custom_metrics_path = configs["custom_metrics"]["metrics_path"]
    original_path = os.path.abspath(os.path.dirname(custom_metrics_path))

    metrics_file_name = os.path.basename(custom_metrics_path)
    custom_metrics_path = os.path.join(project_path, "custom_configs", "metrics")

    configs["custom_metrics"]["metrics_path"] = f"custom_configs/metrics/{metrics_file_name}"

    return original_path, custom_metrics_path


def get_resource(configs: dict, resource_name: str):
    default_resource_file = os.path.join(project_path, "resources", f"{resource_name}.yaml")
    with open(default_resource_file, "r") as f:
        resources = yaml.safe_load(f)

    current_path = os.getcwd()
    if configs.get("custom_resource_dir") is not None:
        custom_resource_dir = configs["custom_resource_dir"]
        resource_file = os.path.join(current_path, custom_resource_dir, f"{resource_name}.yaml")
        if os.path.exists(resource_file):
            with open(resource_file, "r") as f:
                resources = yaml.safe_load(f)
    return resources


def get_resource_constraint(configs: dict):
    constraint_name = configs.get("constraint", "test")
    constraints = get_resource(configs=configs, resource_name="multimodal_constraints")
    constraint_configs = constraints[constraint_name]
    return constraint_configs


def get_framework_configs(configs: dict):
    framework_name = configs.get("framework", "stable")
    frameworks = get_resource(configs=configs, resource_name="multimodal_frameworks")
    framework_configs = frameworks[framework_name]
    return framework_configs


@app.command()
def run(
    config_file: str = typer.Argument(..., help="Path to custom config file."),
    remove_resources: bool = typer.Option(False, help="Remove resources after run."),
    wait: bool = typer.Option(False, help="Whether to block and wait for the benchmark to finish, default to False."),
    skip_setup: bool = typer.Option(
        False, help="Whether to skip setting up framework in local mode, default to False."
    ),
    save_hardware_metrics: bool = typer.Option(False, help="Whether to query and save the hardware metrics."),
):
    """Main function that runs the benchmark based on the provided configuration options."""
    configs = {}
    if config_file.startswith("s3"):
        config_file = download_file_from_s3(s3_path=config_file)
    with open(config_file, "r") as f:
        configs = yaml.safe_load(f)
        if isinstance(configs, list):
            configs = configs[
                int(os.environ.get("AWS_BATCH_JOB_ARRAY_INDEX", "0"))
            ]  # AWS array job sets AWS_BATCH_JOB_ARRAY_INDEX for child jobs

    benchmark_name = _get_benchmark_name(configs=configs)
    timestamp_pattern = r"\d{8}T\d{6}"  # Timestamp that matches YYYYMMDDTHHMMSS
    if not re.search(timestamp_pattern, benchmark_name):
        benchmark_name += "_" + formatted_time()

    root_dir = configs.get("root_dir", "ag_bench_runs")
    module = configs["module"]
    benchmark_dir = os.path.join(root_dir, module, benchmark_name)
    tmpdir = None

    if configs["mode"] == "aws":
        paths = []

        os_name = platform.system()
        if os_name != "Linux":
            raise NotImplementedError("Only Linux system is supported to run AWS mode at the moment.")

        try:
            configs["benchmark_name"] = benchmark_name
            # setting environment variables for docker build ARG
            os.environ["AG_BENCH_VERSION"] = agbench_version

            os.environ["FRAMEWORK_PATH"] = f"frameworks/{module}"

            if module in AMLB_DEPENDENT_MODULES:
                os.environ["AMLB_FRAMEWORK"] = configs["framework"]
                os.environ["GIT_URI"], os.environ["GIT_BRANCH"] = _get_git_info(configs["git_uri#branch"])

                if configs.get("amlb_constraint") is None:
                    configs["amlb_constraint"] = "test"

                amlb_user_dir = configs.get("amlb_user_dir")
                tmpdir = None
                if amlb_user_dir is not None:
                    if amlb_user_dir.startswith("s3://"):
                        tmpdir = tempfile.TemporaryDirectory()
                        amlb_user_dir_local = download_dir_from_s3(s3_path=amlb_user_dir, local_path=tmpdir.name)
                    else:
                        amlb_user_dir_local = amlb_user_dir

                    default_user_dir = "custom_configs/amlb_configs"
                    custom_configs_path = os.path.join(project_path, default_user_dir)
                    lambda_custom_configs_path = os.path.join(
                        project_path, "cloud/aws/batch_stack/lambdas", default_user_dir
                    )
                    original_path = amlb_user_dir_local
                    paths += [custom_configs_path, lambda_custom_configs_path]
                    for path in paths:
                        # mounting custom directory to a predefined directory in the package
                        # to make it available for Docker build
                        _umount_if_needed(path)
                        _mount_dir(orig_path=original_path, new_path=path)
                    os.environ["AMLB_USER_DIR"] = default_user_dir  # For Docker build
                    configs["amlb_user_dir"] = default_user_dir  # For Lambda job config
            elif module in INDEPENDENT_MODULES:
                if configs.get("custom_dataloader") is not None:
                    original_path, custom_dataloader_path = update_custom_dataloader(configs=configs)
                    paths.append(custom_dataloader_path)
                    _umount_if_needed(custom_dataloader_path)
                    _mount_dir(orig_path=original_path, new_path=custom_dataloader_path)

                if configs.get("custom_metrics") is not None:
                    original_path, custom_metrics_path = update_custom_metrics(configs=configs)
                    paths.append(custom_metrics_path)
                    _umount_if_needed(custom_metrics_path)
                    _mount_dir(orig_path=original_path, new_path=custom_metrics_path)

                resource_constraint = get_resource_constraint(configs=configs)
                configs["cdk_context"].update(resource_constraint)

                framework_configs = get_framework_configs(configs=configs)

                if configs.get("custom_resource_dir") is not None:
                    custom_resource_path = os.path.join(project_path, "custom_configs", "resources")
                    paths.append(custom_resource_path)
                    _umount_if_needed(custom_resource_path)
                    _mount_dir(orig_path=configs["custom_resource_dir"], new_path=custom_resource_path)
                    configs["custom_resource_dir"] = "custom_configs/resources"

                os.environ["GIT_URI"] = framework_configs["repo"]
                os.environ["GIT_BRANCH"] = framework_configs.get("version", "stable")

            infra_configs = deploy_stack(custom_configs=configs)

            cloud_config_path = _dump_configs(
                benchmark_dir=benchmark_dir, configs=configs, file_name=os.path.basename(config_file)
            )
            config_s3_path = upload_to_s3(
                s3_bucket=infra_configs["METRICS_BUCKET"],
                s3_dir=f"configs/{module}/{benchmark_name}",
                local_path=cloud_config_path,
            )

            response = invoke_lambda(configs=infra_configs, config_file=config_s3_path)

            job_configs = {"job_configs": response}
            aws_configs = {**infra_configs, **job_configs}
            logger.info(f"Saving infra configs and submitted job configs under {benchmark_dir}.")
            aws_config_path = _dump_configs(
                benchmark_dir=benchmark_dir, configs=aws_configs, file_name="aws_configs.yaml"
            )

            if remove_resources:
                wait = True
            if wait or save_hardware_metrics:
                logger.info(
                    "Waiting for jobs to complete. You can quit at anytime and the benchmark will continue to run on the cloud"
                )
                if remove_resources:
                    logger.info(
                        "Resources will be deleted after the jobs are finished. You can also call \n"
                        f"`agbench destroy-stack --config-file {aws_config_path}` "
                        "to delete the stack after jobs have run to completion if you choose to quit now."
                    )
                time.sleep(10)  # wait for Batch to load
                failed_jobs = wait_for_jobs(config_file=aws_config_path)
                if len(failed_jobs) > 0:
                    logger.warning("Some jobs have failed: %s.", failed_jobs)
                    if remove_resources:
                        logger.warning("Resources will be kept so error logs can be accessed")
                elif failed_jobs is None:
                    if remove_resources:
                        logger.error("Resources are not being removed due to errors.")
                else:
                    logger.info("All job succeeded.")
                    if save_hardware_metrics:
                        get_hardware_metrics(
                            config_file=aws_config_path,
                            s3_bucket=infra_configs["METRICS_BUCKET"],
                            module=module,
                            benchmark_name=benchmark_name,
                        )
                    if remove_resources:
                        logger.info("Removing resoureces...")
                        destroy_stack(
                            static_resource_stack=infra_configs["STATIC_RESOURCE_STACK_NAME"],
                            batch_stack=infra_configs["BATCH_STACK_NAME"],
                            cdk_deploy_account=infra_configs["CDK_DEPLOY_ACCOUNT"],
                            cdk_deploy_region=infra_configs["CDK_DEPLOY_REGION"],
                            config_file=None,
                        )
                        logger.info("Resources removed successfully.")
        finally:
            for path in paths:
                _umount_if_needed(path)

    elif configs["mode"] == "local":
        split_id = os.environ.get("AWS_BATCH_JOB_ARRAY_INDEX", 0)
        benchmark_dir_s3 = f"{module}/{benchmark_name}"
        if split_id is not None:
            benchmark_dir_s3 += f"/{benchmark_name}_{split_id}"

        if module in AMLB_DEPENDENT_MODULES:
            amlb_user_dir = configs.get("amlb_user_dir")
            if amlb_user_dir and amlb_user_dir.startswith("s3://"):
                tmpdir = tempfile.TemporaryDirectory()
                configs["amlb_user_dir"] = download_dir_from_s3(s3_path=amlb_user_dir, local_path=tmpdir.name)
        elif module in INDEPENDENT_MODULES:
            resource_constraint = get_resource_constraint(configs=configs)
            configs["time_limit"] = resource_constraint.get("TIME_LIMIT")

        logger.info(f"Running benchmark {benchmark_name} at {benchmark_dir}.")

        run_benchmark(
            benchmark_name=benchmark_name,
            benchmark_dir=benchmark_dir,
            configs=configs,
            benchmark_dir_s3=benchmark_dir_s3,
            skip_setup=skip_setup,
        )
    else:
        raise NotImplementedError

    if tmpdir:
        tmpdir.cleanup()



import abc
import logging
import os

import pandas as pd

from autogluon.bench.utils.dataset_utils import get_data_home_dir, get_repo_url, path_expander
from autogluon.common.loaders import load_zip
from autogluon.common.loaders._utils import download

from .constants import (
    _BINARY,
    _CATEGORICAL,
    _IMAGE_SIMILARITY,
    _IMAGE_TEXT_SIMILARITY,
    _MULTICLASS,
    _NER,
    _NUMERICAL,
    _REGRESSION,
    _TEXT,
    _TEXT_SIMILARITY,
)

# Add dataset class names here
__all__ = [
    "Shopee",
    "StanfordOnline",
    "Flickr30k",
    "SNLI",
    "MitMovies",
    "WomenClothingReview",
    "MelBourneAirBnb",
    "AEPricePrediction",
    "IMDBGenrePrediction",
    "JCPennyCategory",
    "NewsPopularity",
    "NewsChannel",
]

logger = logging.getLogger(__name__)


class BaseMultiModalDataset(abc.ABC):
    def __init__(self, split: str, dataset_name: str, data_info: dict):
        """
        Initializes the class.

        Args:
            split (str): Specifies the dataset split. It should be one of the following options: 'train', 'val', 'test'.
        """
        try:
            ext = os.path.splitext(data_info[split]["url"])[-1]
            self._path = os.path.join(get_data_home_dir(), dataset_name, f"{split}{ext}")
            download(data_info[split]["url"], path=self._path)
            if ext == ".csv":
                self._data = pd.read_csv(self._path)
            elif ext == ".pq":
                self._data = pd.read_parquet(self._path)
            else:
                raise NotImplementedError(f"File extension {ext} is not supported.")
        except Exception:
            logger.warn(f"The data split {split} is not available.")
            self._data = None

        self._split = split

    @property
    @abc.abstractmethod
    def feature_columns(self):
        pass

    @property
    @abc.abstractmethod
    def label_columns(self):
        pass

    @property
    def label_types(self):
        pass

    @property
    @abc.abstractmethod
    def data(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def metric(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def problem_type(self):
        raise NotImplementedError


class BaseImageDataset(BaseMultiModalDataset):
    @property
    def base_folder(self):
        """Base folder that contains images"""
        pass

    @property
    @abc.abstractmethod
    def image_columns(self):
        """List of image columns"""
        pass


class BaseMatcherDataset(BaseMultiModalDataset):
    @property
    def match_label(self):
        """the label indicating that query and response have the same semantic meanings"""
        pass


class Shopee(BaseImageDataset):
    _SOURCE = ""
    _INFO = {
        "data": {
            "url": get_repo_url() + "vision_datasets/shopee.zip",
        },
    }
    _registry_name = "shopee"

    def __init__(self, split="train"):
        self._split = split
        self._path = os.path.join(get_data_home_dir(), "shopee")
        load_zip.unzip(self._INFO["data"]["url"], unzip_dir=self._path)
        self._base_folder = os.path.join(self._path, "shopee")
        try:
            data_path = os.path.join(self._base_folder, f"{self._split}.csv")
            self._data = pd.read_csv(data_path)
            self._data["image"] = self._data["image"].apply(
                lambda ele: path_expander(ele, base_folder=self._base_folder)
            )
        except FileNotFoundError as e:
            logger.warn(f"The data split {self._split} is not available.")
            self._data = None

    @property
    def base_folder(self):
        """Base folder that contains images"""
        return self._base_folder

    @property
    def image_columns(self):
        """List of image columns"""
        return ["image"]

    @property
    def data(self):
        return self._data

    @property
    def feature_columns(self):
        return ["image"]

    @property
    def label_columns(self):
        return ["label"]

    @property
    def metric(self):
        return "acc"

    @property
    def problem_type(self):
        return _MULTICLASS


class StanfordOnline(BaseMatcherDataset):
    _SOURCE = "https://cvgl.stanford.edu/projects/lifted_struct/"
    _INFO = {
        "data": {
            "url": get_repo_url() + "Stanford_Online_Products.zip",
        },
    }
    _registry_name = "stanford_online"

    def __init__(self, split="train"):
        self._split = split
        self._path = os.path.join(get_data_home_dir(), "Stanford_Online_Products")
        load_zip.unzip(self._INFO["data"]["url"], unzip_dir=self._path)
        self._base_folder = os.path.join(self._path, "Stanford_Online_Products")
        try:
            self._data = pd.read_csv(os.path.join(self._base_folder, f"{self._split}.csv"), index_col=0)
            self._image_columns = ["Image1", "Image2"]
            for image_col in self._image_columns:
                self._data[image_col] = self._data[image_col].apply(
                    lambda ele: path_expander(ele, base_folder=self._base_folder)
                )
        except FileNotFoundError as e:
            logger.warn(f"The data split {self._split} is not available.")
            self._data = None

    @property
    def image_columns(self):
        """List of image columns"""
        return self._image_columns

    @property
    def data(self):
        return self._data

    @property
    def feature_columns(self):
        return self._image_columns

    @property
    def label_columns(self):
        return ["Label"]

    @property
    def match_label(self):
        return 1

    @property
    def metric(self):
        return "roc_auc"

    @property
    def problem_type(self):
        return _IMAGE_SIMILARITY


class Flickr30k(BaseMatcherDataset):
    _SOURCE = "https://paperswithcode.com/dataset/flickr30k"
    _INFO = {
        "data": {"url": get_repo_url() + "flickr30k.zip"},
    }
    _registry_name = "flickr30k"

    def __init__(self, split="train"):
        self._split = split
        self._path = os.path.join(get_data_home_dir(), "flickr30k")
        load_zip.unzip(self._INFO["data"]["url"], unzip_dir=self._path)
        self._base_folder = os.path.join(self._path, "flickr30k_processed")

        try:
            self._data = pd.read_csv(os.path.join(self._base_folder, f"{self._split}.csv"), index_col=0)
            self._image_col = "image"
            self._text_col = "caption"

            self._data[self._image_col] = self._data[self._image_col].apply(
                lambda ele: path_expander(ele, base_folder=self._base_folder)
            )
            self._label_col = "relevance"
            self._data[self._label_col] = [1] * len(self._data)
        except FileNotFoundError as e:
            logger.warn(f"The data split {self._split} is not available.")
            self._data = None

    @property
    def image_columns(self):
        """List of image columns"""
        return [self._image_col]

    @property
    def text_columns(self):
        """List of text columns"""
        return [self._text_col]

    @property
    def data(self):
        return self._data

    @property
    def feature_columns(self):
        return [self._image_col, self._text_col]

    @property
    def label_columns(self):
        return [self._label_col]

    @property
    def match_label(self):
        return 1

    @property
    def metric(self):
        return "recall"

    @property
    def problem_type(self):
        return _IMAGE_TEXT_SIMILARITY


class SNLI(BaseMatcherDataset):
    _SOURCE = "https://nlp.stanford.edu/projects/snli/"
    _INFO = {
        "train": {
            "url": get_repo_url() + "snli/snli_train.csv",
        },
        "test": {"url": get_repo_url() + "snli/snli_test.csv"},
    }
    _registry_name = "snli"

    def __init__(self, split="train"):
        self._split = split
        self._path = os.path.join(get_data_home_dir(), "snli", f"{split}.csv")
        try:
            download(self._INFO[split]["url"], path=self._path)
            self._data = pd.read_csv(self._path, delimiter="|")
        except Exception:
            logger.warn(f"The data split {self._split} is not available.")
            self._data = None

    @property
    def text_columns(self):
        """List of text columns"""
        return ["premise", "hypothesis"]

    @property
    def data(self):
        return self._data

    @property
    def feature_columns(self):
        return ["premise", "hypothesis"]

    @property
    def label_columns(self):
        return ["label"]

    @property
    def match_label(self):
        return 1

    @property
    def metric(self):
        return "roc_auc"

    @property
    def problem_type(self):
        return _TEXT_SIMILARITY


class MitMovies(BaseMultiModalDataset):
    _SOURCE = "https://groups.csail.mit.edu/sls/downloads/movie/"
    _INFO = {
        "train": {
            "url": get_repo_url() + "ner/mit-movies/train_v2.csv",
        },
        "test": {
            "url": get_repo_url() + "ner/mit-movies/test_v2.csv",
        },
    }
    _registry_name = "mit_movies"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @property
    def feature_columns(self):
        return ["text_snippet"]

    @property
    def label_columns(self):
        return ["entity_annotations"]

    @property
    def metric(self):
        return ["overall_recall", "overall_precision", "overall_f1", "actor"]

    @property
    def problem_type(self):
        return _NER

    @property
    def data(self):
        return self._data


class WomenClothingReview(BaseMultiModalDataset):
    _SOURCE = "https://www.kaggle.com/nicapotato/womens-ecommerce-clothing-reviews"
    _INFO = {
        "train": {
            "url": get_repo_url() + "women_clothing_review/train.pq",
        },
        "test": {
            "url": get_repo_url() + "women_clothing_review/test.pq",
        },
    }
    _registry_name = "women_clothing_review"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def feature_columns(self):
        return ["Title", "Review Text", "Age", "Division Name", "Department Name", "Class Name"]

    @property
    def feature_types(self):
        return [_TEXT, _TEXT, _NUMERICAL, _CATEGORICAL, _CATEGORICAL, _CATEGORICAL]

    @property
    def fill_na_value(self):
        """The default function to fill missing values"""
        return {"Division Name": "None", "Department Name": "None", "Class Name": "None"}

    @property
    def label_columns(self):
        return ["Rating"]

    @property
    def label_types(self):
        return [_NUMERICAL]

    @property
    def data(self):
        return self._data

    @property
    def metric(self):
        return "r2"

    @property
    def problem_type(self):
        return _REGRESSION


class MelBourneAirBnb(BaseMultiModalDataset):
    _SOURCE = "https://www.kaggle.com/tylerx/melbourne-airbnb-open-data"
    _INFO = {
        "train": {
            "url": get_repo_url() + "airbnb_melbourne/train.pq",
        },
        "test": {
            "url": get_repo_url() + "airbnb_melbourne/test.pq",
        },
    }
    _registry_name = "melbourne_airbnb"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def ignore_columns(self):
        return [
            "id",
            "listing_url",
            "scrape_id",
            "last_scraped",
            "picture_url",
            "host_id",
            "host_url",
            "host_name",
            "host_thumbnail_url",
            "host_picture_url",
            "monthly_price",
            "weekly_price",
            "price",
            "calendar_last_scraped",
        ]

    @property
    def label_columns(self):
        return ["price_label"]

    @property
    def feature_columns(self):
        all_columns = sorted(self._data.columns)
        feature_columns = [
            col for col in all_columns if col not in self.label_columns and col not in self.ignore_columns
        ]
        return feature_columns

    @property
    def data(self):
        return self._data

    @property
    def label_types(self):
        return [_CATEGORICAL]

    @property
    def metric(self):
        return "acc"

    @property
    def problem_type(self):
        return _MULTICLASS


class AEPricePrediction(BaseMultiModalDataset):
    _SOURCE = "https://www.kaggle.com/PromptCloudHQ/innerwear-data-from-victorias-secret-and-others"
    _INFO = {
        "train": {
            "url": get_repo_url() + "ae_price_prediction/train.pq",
        },
        "test": {
            "url": get_repo_url() + "ae_price_prediction/test.pq",
        },
    }
    _registry_name = "ae_price_prediction"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def data(self):
        return self._data

    @property
    def ignore_columns(self):
        return ["mrp", "pdp_url"]

    @property
    def feature_columns(self):
        return [col for col in self.data.columns if col not in self.ignore_columns and col not in self.label_columns]

    @property
    def label_columns(self):
        return ["price"]

    @property
    def label_types(self):
        return [_NUMERICAL]

    @property
    def metric(self):
        return "r2"

    @property
    def problem_type(self):
        return _REGRESSION


class IMDBGenrePrediction(BaseMultiModalDataset):
    _SOURCE = "https://www.kaggle.com/PromptCloudHQ/imdb-data"
    _INFO = {
        "train": {
            "url": get_repo_url() + "imdb_genre_prediction/train.csv",
        },
        "test": {
            "url": get_repo_url() + "imdb_genre_prediction/test.csv",
        },
    }
    _registry_name = "imdb_genre_prediction"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @property
    def data(self):
        return self._data

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def data(self):
        return self._data

    @property
    def label_columns(self):
        return ["Genre_is_Drama"]

    @property
    def label_types(self):
        return [_CATEGORICAL]

    @property
    def feature_columns(self):
        return [col for col in list(self.data.columns) if col not in self.label_columns]

    @property
    def metric(self):
        return "roc_auc"

    @property
    def problem_type(self):
        return _BINARY


class JCPennyCategory(BaseMultiModalDataset):
    _SOURCE = "https://www.kaggle.com/PromptCloudHQ/all-jc-penny-products"
    _INFO = {
        "train": {
            "url": get_repo_url() + "jc_penney_products/train.csv",
        },
        "test": {
            "url": get_repo_url() + "jc_penney_products/test.csv",
        },
    }
    _registry_name = "jc_penney_products"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @property
    def data(self):
        return self._data

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def data(self):
        return self._data

    @property
    def label_columns(self):
        return ["sale_price"]

    @property
    def label_types(self):
        return [_NUMERICAL]

    @property
    def feature_columns(self):
        return [col for col in list(self.data.columns) if col not in self.label_columns]

    @property
    def metric(self):
        return "r2"

    @property
    def problem_type(self):
        return _REGRESSION


class NewsPopularity(BaseMultiModalDataset):
    _SOURCE = "https://archive.ics.uci.edu/ml/datasets/online+news+popularity"
    _INFO = {
        "train": {
            "url": get_repo_url() + "news_popularity2/train.csv",
        },
        "test": {
            "url": get_repo_url() + "news_popularity2/test.csv",
        },
    }
    _registry_name = "news_popularity"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @property
    def data(self):
        return self._data

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def data(self):
        return self._data

    @property
    def label_columns(self):
        return ["log_shares"]

    @property
    def label_types(self):
        return [_NUMERICAL]

    @property
    def feature_columns(self):
        return [col for col in list(self.data.columns) if col not in self.label_columns]

    @property
    def metric(self):
        return "r2"

    @property
    def problem_type(self):
        return _REGRESSION


class NewsChannel(BaseMultiModalDataset):
    _SOURCE = "https://archive.ics.uci.edu/ml/datasets/online+news+popularity"
    _INFO = {
        "train": {
            "url": get_repo_url() + "news_channel/train.csv",
        },
        "test": {
            "url": get_repo_url() + "news_channel/test.csv",
        },
    }
    _registry_name = "news_channel"

    def __init__(self, split="train"):
        super().__init__(split=split, dataset_name=self._registry_name, data_info=self._INFO)

    @property
    def data(self):
        return self._data

    @classmethod
    def splits(cls):
        return cls._INFO.keys()

    @property
    def data(self):
        return self._data

    @property
    def label_columns(self):
        return ["channel"]

    @property
    def label_types(self):
        return [_CATEGORICAL]

    @property
    def feature_columns(self):
        return [col for col in list(self.data.columns) if col not in self.label_columns]

    @property
    def metric(self):
        return "acc"

    @property
    def problem_type(self):
        return _MULTICLASS



import csv
import logging
import os
import tempfile
from datetime import datetime, timedelta
from typing import List, Optional

import boto3
import pandas as pd
import typer
import yaml

from autogluon.bench.utils.general_utils import upload_to_s3

aws_account_id = None
aws_account_region = None

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def find_s3_file(s3_bucket: str, prefix: str, file: str):
    s3 = boto3.client("s3")
    paginator = s3.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=s3_bucket, Prefix=prefix)

    for page in page_iterator:
        if "Contents" in page:
            for obj in page["Contents"]:
                if obj["Key"].endswith("results.csv"):
                    return f"s3://{s3_bucket}/{obj['Key']}"
    return None


def get_job_ids(config: dict):
    """
    This function returns a list of job IDs of all jobs ran for a benchmark run
    Parameters
    ----------
    config_file: str,
        Path to config file containing job IDs
    """
    job_ids = list(config.get("job_configs", {}).keys())
    return job_ids


def get_instance_id(job_id):
    """
    This function returns the instance ID (ARN) of the EC2 instance that was used to run a job with given job ID.
    Parameters
    ----------
    job_id: str
    """
    batch_client = boto3.client("batch", region_name=aws_account_region)
    ecs_client = boto3.client("ecs", region_name=aws_account_region)

    response = batch_client.describe_jobs(jobs=[f"{job_id}"])
    if response:
        container_arn = response["jobs"][0]["container"]["containerInstanceArn"]
        cluster_arn = response["jobs"][0]["container"]["taskArn"].split("/")
        cluster = f"arn:aws:ecs:{aws_account_region}:{aws_account_id}:cluster/" + cluster_arn[1]

    response = ecs_client.describe_container_instances(cluster=cluster, containerInstances=[container_arn])
    instance_id = response["containerInstances"][0]["ec2InstanceId"]
    return instance_id


def get_instance_util(
    namespace: str,
    instance_id: str,
    metric: str,
    start_time: datetime,
    end_time: datetime,
    cloudwatch_client: boto3.client,
    period: int = 360,
    statistics: Optional[List[str]] = ["Average"],
) -> dict:
    """
    This function returns the instance ID of the EC2 instance that was used to run a job with given job ID.
    Refer to https://docs.aws.amazon.com/cli/latest/reference/cloudwatch/get-metric-statistics.html for docs on how to interact with the CloudWatch API
    Also refer to https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch/client/get_metric_statistics.html for documentation on how to interact with the API through Python
    Parameters
    ----------
    instance_id: str,
        EC2 instance ARN
    metric: str,
        Name of metric to pass into the CloudWatch API. Example: CPUUtilization
        Refer to https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/viewing_metrics_with_cloudwatch.html#ec2-cloudwatch-metrics
    start_time: datetime,
    end_time: datetime,
    statistics: Optional[List[str]] = ["Average"],
        The metric statistics, other than percentile. For percentile statistics, use `ExtendedStatistics` . When calling `get_metric_statistics` , you must specify either `Statistics` or `ExtendedStatistics` , but not both.
        Examples: Average, Maximum, Minimum
    """
    return cloudwatch_client.get_metric_statistics(
        Namespace=namespace,
        MetricName=metric,
        Dimensions=[
            {"Name": "InstanceId", "Value": instance_id},
        ],
        Statistics=statistics,
        StartTime=start_time,
        EndTime=end_time,
        Period=period,
    )


def format_metrics(
    instance_metrics: dict,
    framework: str,
    dataset: str,
    fold: int,
    mode: str,
    statistics: Optional[List[str]] = ["Average"],
):
    """
    This function returns a formatted version of the dictionary of metrics provided by the CloudWatch API so it can be easily added to a CSV file and passed into `autogluon-dashboard`.
    Parameters
    ----------
    instance_metrics: dict,
        Dictionary of instance metrics for a given EC2 instance provided by CloudWatch
    framework: str,
        Name of the framework
    dataset: str,
        Name of the dataset
    fold: int,
        Fold #
    mode: str,
        Mode -> Training or Prediction
    statistics: Optional[List[str]] = ["Average"],
        The metric statistics, other than percentile. For percentile statistics, use `ExtendedStatistics` . When calling `get_metric_statistics` , you must specify either `Statistics` or `ExtendedStatistics` , but not both.
        Examples: Average, Maximum, Minimum
    """
    output_dict = {}
    output_dict["framework"] = framework
    output_dict["dataset"] = dataset
    output_dict["mode"] = mode
    output_dict["fold"] = fold
    output_dict["metric"] = instance_metrics["Label"]
    for i in range(len(instance_metrics["Datapoints"])):
        for stat in statistics:
            output_dict["framework"] = framework
            output_dict["dataset"] = dataset
            output_dict["mode"] = mode
            output_dict["fold"] = fold
            output_dict["metric"] = instance_metrics["Label"]
            output_dict["statistic_type"] = stat
            output_dict["statistic_value"] = instance_metrics["Datapoints"][i][f"{stat}"]
            output_dict["unit"] = instance_metrics["Datapoints"][i]["Unit"]
    return output_dict


def get_metrics(
    job_id: str,
    s3_bucket: str,
    module: str,
    benchmark_name: str,
    sub_folder: str,
    cloudwatch_client: boto3.client,
    namespace: str = "EC2",  # CloudWatch "Custom" namespace, i.e. Custom/EC2
):
    """
    Parameters
    ----------
    job_id: str,
    metrics: list,
        List of metrics to pass into the CloudWatch API. Example: CPUUtilization
        Refer to https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/viewing_metrics_with_cloudwatch.html#ec2-cloudwatch-metrics
    s3_bucket: str,
    module: str,
    benchmark_name: str,
    sub_folder: str,
        Sub folder for results.csv file.
        Passed in from `get_hardware_metrics` function
    namespace: str,
        CloudWatch Metrics Namespace, default: AWS/EC2
    """
    path_prefix = f"{module}/{benchmark_name}/{sub_folder}/"
    s3_path_to_csv = find_s3_file(s3_bucket=s3_bucket, prefix=path_prefix, file="results.csv")
    results = pd.read_csv(s3_path_to_csv)
    metrics_list = []
    instance_id = get_instance_id(job_id)
    metrics_data = cloudwatch_client.list_metrics(
        Dimensions=[
            {"Name": "InstanceId", "Value": instance_id},
        ],
        Namespace=namespace,
    )["Metrics"]
    metrics_pool = [item["MetricName"] for item in metrics_data]

    for metric in metrics_pool:
        for i in results.index:
            framework, dataset, utc, train_time, predict_time, fold = (
                results["framework"][i],
                results["task"][i],
                results["utc"][i],
                results["training_duration"][i],
                results["predict_duration"][i],
                results["fold"][i],
            )
            utc_dt = datetime.strptime(utc, "%Y-%m-%dT%H:%M:%S")
            period = int((timedelta(seconds=train_time) + timedelta(seconds=predict_time)).total_seconds())
            if period < 60:
                period = 60
            elif period % 60 != 0:
                period = (period // 60) * 60  # Round down to the nearest multiple of 60

            training_util = get_instance_util(
                namespace=namespace,
                instance_id=instance_id,
                metric=metric,
                start_time=utc_dt,
                end_time=utc_dt + timedelta(seconds=train_time) + timedelta(seconds=predict_time),
                period=period,
                cloudwatch_client=cloudwatch_client,
            )
            predict_util = get_instance_util(
                namespace=namespace,
                instance_id=instance_id,
                metric=metric,
                start_time=utc_dt - timedelta(minutes=predict_time),
                end_time=utc_dt,
                period=period,
                cloudwatch_client=cloudwatch_client,
            )
            if training_util["Datapoints"]:
                metrics_list.append(format_metrics(training_util, framework, dataset, fold, "Training"))
            if predict_util["Datapoints"]:
                metrics_list.append(format_metrics(predict_util, framework, dataset, fold, "Prediction"))
    return metrics_list


def save_results(metrics_list: list, path: str):
    """
    Writes the formatted dictionary of metrics to a csv to pass into `autogluon-dashboard`.
    Parameters
    ----------
    metrics_list: list,
        List of hardware metrics to write to CSV
    path: str:
        Path to save file
    """
    csv_headers = ["framework", "dataset", "mode", "fold", "metric", "statistic_type", "statistic_value", "unit"]
    csv_location = os.path.join(path, "hardware_metrics.csv")
    with open(csv_location, "w", newline="") as csvFile:
        writer = csv.DictWriter(csvFile, fieldnames=csv_headers)
        writer.writeheader()
        writer.writerows(metrics_list)
    return csv_location


def get_hardware_metrics(
    config_file: str = typer.Argument(help="Path to YAML config file containing job ids."),
    s3_bucket: str = typer.Argument(help="Name of the S3 bucket to which the benchmark results were outputted."),
    module: str = typer.Argument(help="Can be one of ['tabular', 'timeseries', 'multimodal']."),
    benchmark_name: str = typer.Argument(
        help="Folder name of benchmark run in which all objects with path 'scores/results.csv' get aggregated."
    ),
):
    """
    External API function to interact with the script.
    Parameters
    ----------
    config_file: str,
        Path to config file containing job IDs
    s3_bucket: str,
        Name of the S3 bucket to which the benchmark results were outputted.
    module: str,
        Benchmark module: tabular or multimodal
    benchmark_name: str,
        Name of the benchmark
        Example: ag_bench_20230817T123456
    """
    if not config_file:
        raise ValueError("Invalid Config File")
    logger.info(f"Getting hardware metrics for jobs under config file: {config_file}")
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)
    job_ids = get_job_ids(config)

    global aws_account_id, aws_account_region
    aws_account_id = config.get("CDK_DEPLOY_ACCOUNT")
    aws_account_region = config.get("CDK_DEPLOY_REGION")

    cloudwatch_client = boto3.client("cloudwatch", region_name=aws_account_region)
    batch_client = boto3.client("batch", region_name=aws_account_region)

    metrics_list = []
    for job_id in job_ids:
        response = batch_client.describe_jobs(jobs=[job_id])
        job_detail = response["jobs"][0]

        # Check if the job is an array job
        if "arrayProperties" in job_detail and "size" in job_detail["arrayProperties"]:
            size = job_detail["arrayProperties"]["size"]
            sub_ids = [f"{job_id}:{i}" for i in range(size)]
        else:
            sub_ids = [job_id]

        for sub_id in sub_ids:
            id = sub_id.split(":")[-1]
            if id != "":
                id = "_" + id
            sub_folder = f"{benchmark_name}{id}"
            metrics_list += get_metrics(
                job_id=sub_id,
                s3_bucket=s3_bucket,
                module=module,
                benchmark_name=benchmark_name,
                sub_folder=sub_folder,
                cloudwatch_client=cloudwatch_client,
            )
    if metrics_list:
        with tempfile.TemporaryDirectory() as temp_dir:
            local_path = save_results(metrics_list, temp_dir)
            upload_to_s3(s3_bucket=s3_bucket, s3_dir=f"{module}/{benchmark_name}", local_path=local_path)



from autogluon.common.loaders import load_s3
from autogluon.common.utils import s3_utils


def get_s3_paths(path_prefix: str, contains=None, suffix=None):
    bucket, prefix = s3_utils.s3_path_to_bucket_prefix(path_prefix)
    objects = load_s3.list_bucket_prefix_suffix_contains_s3(
        bucket=bucket, prefix=prefix, suffix=suffix, contains=contains
    )
    paths_full = [s3_utils.s3_bucket_prefix_to_path(bucket=bucket, prefix=file, version="s3") for file in objects]
    return paths_full



from . import multimodal_dataset, object_detection_dataset
from .registry import Registry

multimodal_dataset_registry = Registry("multimodal_dataset")

for data_classes in [multimodal_dataset, object_detection_dataset]:
    for class_name in data_classes.__all__:
        dataset_class = getattr(data_classes, class_name)
        multimodal_dataset_registry.register(dataset_class._registry_name, dataset_class)



from __future__ import annotations

import logging
import os
from typing import Dict, List, Optional

import pandas as pd
import typer
from typing_extensions import Annotated

from autogluon.bench.eval.evaluation.constants import (
    DATASET,
    FOLD,
    FRAMEWORK,
    METRIC,
    METRIC_ERROR,
    PROBLEM_TYPE,
    TIME_INFER_S,
    TIME_TRAIN_S,
)
from autogluon.bench.eval.evaluation.preprocess import preprocess_openml
from autogluon.common.savers import save_pd

app = typer.Typer()
logger = logging.getLogger(__name__)

MINIMAL_COLUMNS = [
    DATASET,
    FOLD,
    FRAMEWORK,
    "constraint",
    METRIC,
    METRIC_ERROR,
    TIME_TRAIN_S,
    TIME_INFER_S,
    PROBLEM_TYPE,
    "tid",
]


@app.command()
def clean_amlb_results(
    benchmark_name: str = typer.Argument(
        None, help="Benchmark name populated by benchmark run, in format <benchmark_name>_<timestamp>"
    ),
    results_dir: str = typer.Option("data/results/", help="Root directory of raw and prepared results."),
    results_dir_input: str = typer.Option(
        None,
        help="Directory of the results file '<file_prefix><constraint_str><benchmark_name_str>.csv' getting cleaned. Can be an S3 URI. If not provided, it defaults to '<results_dir>input/raw/'",
    ),
    results_dir_output: str = typer.Option(
        None,
        help="Output directory of cleaned file. Can be an S3 URI. If not provided, it defaults to '<results_dir>input/prepared/openml/'",
    ),
    file_prefix: str = typer.Option("results_automlbenchmark", help="File prefix of the input results files."),
    benchmark_name_in_input_path: bool = False,
    benchmark_name_in_output_path: bool = True,
    out_path_prefix: str = typer.Option("openml_ag_", help="Prefix of result file."),
    out_path_suffix: str = typer.Option("", help="suffix of result file."),
    framework_suffix_column: str = typer.Option("constraint", help="Framework suffix column."),
):
    """
    Cleans and aggregate results further with unified column names and adds benchmark name into framework column.

    Example:
        agbench clean-and-save-results ag_tabular_20230629T140546 --results-dir-input s3://autogluon-benchmark-metrics/aggregated/tabular/ag_tabular_20230629T140546/ --benchmark-name-in-input-path --constraints constratint_1 --constraints constratint_2
    """
    clean_and_save_results(
        run_name=benchmark_name,
        results_dir=results_dir,
        results_dir_input=results_dir_input,
        results_dir_output=results_dir_output,
        file_prefix=file_prefix,
        run_name_in_input_path=benchmark_name_in_input_path,
        run_name_in_output_path=benchmark_name_in_output_path,
        save=True,
        out_path_prefix=out_path_prefix,
        out_path_suffix=out_path_suffix,
        framework_suffix_column=framework_suffix_column,
    )


def clean_results_df(
    df_raw: pd.DataFrame,
    framework_suffix: str = None,
    framework_suffix_column: str = None,
    framework_rename_dict: Dict[str, str] = None,
) -> pd.DataFrame:
    df_processed = (
        preprocess_openml.preprocess_openml_input(
            df=df_raw,
            framework_suffix=framework_suffix,
            framework_suffix_column=framework_suffix_column,
            framework_rename_dict=framework_rename_dict,
        )
        .sort_values(by=[DATASET, FOLD, FRAMEWORK])
        .reset_index(drop=True)
    )

    df_processed_columns = list(df_processed.columns)
    df_processed_columns = [c for c in df_processed_columns if c in MINIMAL_COLUMNS] + [
        c for c in df_processed_columns if c not in MINIMAL_COLUMNS
    ]
    df_processed = df_processed[df_processed_columns]
    return df_processed


def clean_and_save_results(
    run_name: str,
    results_dir: str = "data/results/",
    results_dir_input: str | None = None,
    results_dir_output: str | None = None,
    file_prefix: str | List[str] = "results_automlbenchmark",
    run_name_in_input_path: bool = True,
    run_name_in_output_path: bool = True,
    save: bool = True,
    save_minimal: bool = True,
    out_path_prefix: str = "openml_ag_",
    out_path_suffix: str = "",
    framework_suffix_column: str = "constraint",
) -> pd.DataFrame:
    if results_dir_input is None:
        results_dir_input = os.path.join(results_dir, "input/raw/")
    if results_dir_output is None:
        results_dir_output = os.path.join(results_dir, "input/prepared/openml/")
    run_name_str = f"_{run_name}" if run_name_in_input_path else ""

    if not isinstance(file_prefix, list):
        file_prefix = [file_prefix]

    results_list = []
    for prefix in file_prefix:
        results_raw_orig = pd.read_csv(os.path.join(results_dir_input, f"{prefix}{run_name_str}.csv"))
        results = clean_results_df(
            df_raw=results_raw_orig,
            framework_suffix=run_name,
            framework_suffix_column=framework_suffix_column,
        )
        results_list.append(results)
    results_raw = pd.concat(results_list, ignore_index=True, sort=True)

    if save:
        if run_name_in_output_path:
            save_path = os.path.join(results_dir_output, f"{out_path_prefix}{run_name}{out_path_suffix}")
        else:
            save_path = os.path.join(results_dir_output, f"{out_path_prefix}{out_path_suffix}")
        save_path_file = f"{save_path}.csv"

        save_pd.save(path=save_path_file, df=results_raw)
        logger.log(30, f"Cleaned results are saved in file: {save_path_file}")
        save_path_file_pq = f"{save_path}.parquet"
        save_pd.save(path=save_path_file_pq, df=results_raw)
        if save_minimal:
            results_raw_minimal = results_raw[MINIMAL_COLUMNS]

            save_path_file_minimum = f"{save_path}_min.csv"
            save_pd.save(path=save_path_file_minimum, df=results_raw_minimal)
            logger.log(30, f"Cleaned results (minimum) are saved in file: {save_path_file_minimum}")
            save_path_file_minimum_pq = f"{save_path}_min.parquet"
            save_pd.save(path=save_path_file_minimum_pq, df=results_raw_minimal)
    return results_raw


if __name__ == "__main__":
    app()



import typer

from autogluon.bench.cloud.aws.stack_handler import destroy_stack
from autogluon.bench.eval.hardware_metrics.hardware_metrics import get_hardware_metrics
from autogluon.bench.eval.scripts.aggregate_amlb_results import aggregate_amlb_results
from autogluon.bench.eval.scripts.run_evaluation_openml import evaluate_amlb_results
from autogluon.bench.eval.scripts.run_generate_clean_openml import clean_amlb_results
from autogluon.bench.runbenchmark import get_job_status, run
from autogluon.bench.scripts.generate_cloud_configs import generate_cloud_config

app = typer.Typer()

app.command()(run)
app.command()(destroy_stack)
app.command()(get_job_status)
app.command()(generate_cloud_config)
app.command()(aggregate_amlb_results)
app.command()(clean_amlb_results)
app.command()(evaluate_amlb_results)
app.command()(get_hardware_metrics)

if __name__ == "__main__":
    app()



from typing import OrderedDict as t_OrderedDict
from unittest.mock import MagicMock

import pytest

from autogluon.bench.datasets.registry import Registry


class MyClass:
    pass


class TestRegistry:
    def test_create_registry(self):
        registry = Registry("test")
        assert registry._name == "test"
        assert isinstance(registry._obj_map, t_OrderedDict)
        assert len(registry._obj_map) == 0

    def test_register(self):
        registry = Registry("test")
        registry.register(MyClass)
        assert len(registry._obj_map) == 1
        assert list(registry._obj_map.keys())[0] == "MyClass"
        assert registry.get("MyClass") == MyClass

    def test_register_with_nickname(self):
        registry = Registry("test")
        registry.register("alias", MyClass)
        assert len(registry._obj_map) == 1
        assert list(registry._obj_map.keys())[0] == "alias"
        assert registry.get("alias") == MyClass

    def test_get_non_existing_class(self):
        registry = Registry("test")
        with pytest.raises(KeyError):
            registry.get("NonExistingClass")

    def test_list_keys(self):
        registry = Registry("test")
        registry.register(MyClass)
        registry.register("alias", MagicMock())
        keys = registry.list_keys()
        assert "MyClass" in keys
        assert "alias" in keys

    def test_create_class(self):
        registry = Registry("test")
        registry.register(MyClass)
        obj = registry.create("MyClass")
        assert isinstance(obj, MyClass)



metrics = [
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Training",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Prediction",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Training",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Prediction",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Training",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Prediction",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Training",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Prediction",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
]



from __future__ import annotations

import copy
import os
from typing import List, Optional, Set

import pandas as pd
import ray
from tqdm import tqdm

from autogluon.bench.eval.benchmark_context.output_context import OutputContext
from autogluon.bench.eval.benchmark_context.utils import get_s3_paths
from autogluon.common.loaders import load_pd

DEFAULT_COLUMNS_TO_KEEP = [
    "id",
    "task",
    "framework",
    "constraint",
    "fold",
    "type",
    "metric",
    "mode",
    "version",
    "params",
    "app_version",
    "utc",
    "seed",
]


class OutputSuiteContext:
    def __init__(
        self,
        path: str,
        paths_extra: List[str] = None,
        contains: Optional[str] = None,
        allowed_tids: Optional[Set[int]] = None,
        columns_to_keep: Optional[List[str]] = None,
        include_infer_speed: bool = False,
        keep_params: bool = True,
        mode: str = "seq",
        max_workers: int = 64,
    ):
        """
        Parameters
        ----------
        path : str
            The S3 path to the output folder of an AMLB run
            Example: "s3://automl-benchmark-ag/ec2/2023_02_27_zs/"
        paths_extra: List[str], default = None
            A list of additional paths to use when generating `self.output_contexts`.
            Additional output contexts will be identified in the same way as done with `path`.
        contains : Optional[str], default = None
            Can be specified to limit the returned outputs.
            For example, by specifying the constraint, such as ".1h8c."
        allowed_tids : Optional[Set[int]], default = None
            If specified, results will be filtered to only OutputContext objects
            that have an OpenML TID within `allowed_tids`.
        columns_to_keep : List[str], default = None
            The list of columns to keep when loading from results files.
            If None, uses DEFAULT_COLUMNS_TO_KEEP
        include_infer_speed : bool, default = False
            Whether to merge infer_speed results when loading results and leaderboard outputs
        keep_params : bool, default = True
            If False, drops the "params" column from leaderboard output to reduce size of artifact
        mode : str, default = 'seq'
            One of ['seq', 'ray']
            If 'ray' is installed, this will be much faster as data loading will be parallelized.
            'seq' (sequential) should work on all systems, but will be slower than 'ray'.
            'seq' will work properly in a debugger, whereas 'ray' is difficult to debug.
        max_workers : int, default = 64
            If `mode` is set to 'ray', this restricts the maximum number of ray workers to this value.
            No effect if the number of logical cores is less than this value.
            Values greater than 64 can lead to strange errors likely due to S3 service throttling.
        """
        self._path = path
        self.contains = contains
        self.allowed_tids = allowed_tids
        if self.allowed_tids is not None:
            assert isinstance(self.allowed_tids, set)
            for tid in self.allowed_tids:
                assert isinstance(tid, int)
        self.output_contexts = self.get_output_contexts(contains=self.contains, paths_extra=paths_extra)
        if len(self.output_contexts) == 0:
            print(f"WARNING: No output contexts found via path={self._path}, contains={self.contains}")
        self.include_infer_speed = include_infer_speed
        self.mode = mode
        self.max_workers = max_workers
        if columns_to_keep is None:
            columns_to_keep = DEFAULT_COLUMNS_TO_KEEP
        columns_to_keep = copy.deepcopy(columns_to_keep)
        self.keep_params = keep_params
        if not keep_params:
            columns_to_keep = [c for c in columns_to_keep if c != "params"]
        self.columns_to_keep = columns_to_keep

    def get_output_contexts(self, contains: str = None, paths_extra: List[str] = None) -> List[OutputContext]:
        """
        Parameters
        ----------
        contains : str, default = None
            Can be specified to limit the returned outputs.
            For example, by specifying the constraint, such as ".1h8c."
        paths_extra : List[str], default = None
            An optional list of paths to get additional contexts.
        """
        output_contexts = self._get_output_contexts(path=self.path, contains=contains)
        if paths_extra is not None:
            for path in paths_extra:
                output_contexts += self._get_output_contexts(path=path, contains=contains)
        return output_contexts

    @staticmethod
    def _get_output_contexts(path: str, contains: str = None) -> List[OutputContext]:
        """
        Parameters
        ----------
        contains : str, default = None
            Can be specified to limit the returned outputs.
            For example, by specifying the constraint, such as ".1h8c."
        """
        paths_to_results = get_s3_paths(path, contains=contains, suffix="scores/results.csv")
        output_contexts = [OutputContext.from_results_path(path=results_path) for results_path in paths_to_results]
        return output_contexts

    @property
    def path(self):
        return self._path

    @property
    def num_contexts(self):
        return len(self.output_contexts)

    def _loop_func(
        self, func, input_list: list, kwargs=None, allow_exception=False, exception_default=None, mode=None
    ) -> list:
        if len(input_list) == 0:
            return []
        if mode is None:
            mode = self.mode
        process_func = _with_ray if mode == "ray" else _with_seq
        return process_func(
            func=func,
            input_list=input_list,
            kwargs=kwargs,
            allow_exception=allow_exception,
            exception_default=exception_default,
            num_cpus=self.max_workers,
        )

    def load_results(self) -> List[pd.DataFrame]:
        return self._loop_func(
            func=OutputContext.load_results,
            input_list=self.output_contexts,
            kwargs=dict(
                include_infer_speed=self.include_infer_speed,
                keep_params=self.keep_params,
                allowed_tids=self.allowed_tids,
            ),
        )

    def load_zeroshot_metadata(self, max_size_mb: float = None, allow_exception=False) -> List[dict]:
        return self._loop_func(
            func=OutputContext.load_zeroshot_metadata,
            input_list=self.output_contexts,
            kwargs=dict(max_size_mb=max_size_mb),
            allow_exception=allow_exception,
        )

    def get_zeroshot_metadata_size_bytes(self, allow_exception=False) -> List[int]:
        return self._loop_func(
            func=OutputContext.get_zeroshot_metadata_size_bytes,
            input_list=self.output_contexts,
            allow_exception=allow_exception,
        )

    def filter_failures(self):
        amlb_info_list = self.get_amlb_info()
        output_contexts_valid = []
        for info, output_context in zip(amlb_info_list, self.output_contexts):
            if info is None:
                output_contexts_valid.append(output_context)
        print(f"Filtered Failures: {len(output_contexts_valid)}/{len(self.output_contexts)} valid")
        self.output_contexts = output_contexts_valid

    def filter(self, filter_lst: List[bool], inplace=True):
        """
        Filter self.output_contexts by a boolean list. Only keep contexts where the boolean is True.
        """
        if not inplace:
            out = copy.deepcopy(self)
        else:
            out = self

        assert len(filter_lst) == len(out.output_contexts)
        out.output_contexts = [
            output_context for output_context, is_valid in zip(out.output_contexts, filter_lst) if is_valid is True
        ]

        return out

    def aggregate_results(self, results_list: List[pd.DataFrame] | None = None) -> pd.DataFrame:
        if results_list is None:
            results_list = self.load_results()
        results_df = pd.concat(results_list, ignore_index=True)
        return results_df

    def load_leaderboards(self) -> List[pd.DataFrame]:
        if self.num_contexts == 0:
            raise AssertionError("Empty output_contexts!")

        kwargs = dict(
            output_contexts=self.output_contexts,
            columns_to_keep=self.columns_to_keep,
            with_infer_speed=self.include_infer_speed,
        )

        # TODO: Migrate to `self._loop_func`
        if self.mode == "seq":
            result = self._aggregate_leaderboards_seq(**kwargs)
        elif self.mode == "ray":
            result = self._aggregate_leaderboards_ray(**kwargs)
        else:
            raise ValueError(f'Unsupported mode "{self.mode}"')
        print(
            f"Successfully loaded {len(result)}/{self.num_contexts} task outputs "
            f"({round(100 * (len(result) / self.num_contexts), 1)}%)..."
        )
        return result

    def aggregate_leaderboards(self, leaderboards_list: List[pd.DataFrame] | None = None) -> pd.DataFrame:
        if leaderboards_list is None:
            leaderboards_list = self.load_leaderboards()
        leaderboards_df = pd.concat(leaderboards_list, ignore_index=True)
        return leaderboards_df

    def load_model_failures(self) -> List[pd.DataFrame]:
        return self._loop_func(
            func=OutputContext.get_model_failures, input_list=self.output_contexts, allow_exception=True
        )

    def load_info_file_sizes(self, sum: bool = False) -> List[pd.DataFrame]:
        return self._loop_func(
            func=OutputContext.get_info_file_sizes,
            input_list=self.output_contexts,
            kwargs={"sum": sum},
            allow_exception=True,
        )

    def aggregate_info_file_sizes(self, sum: bool = False) -> pd.DataFrame:
        info_file_sizes_lst = self.load_info_file_sizes(sum=sum)
        none_count = 0
        for e in info_file_sizes_lst:
            if e is None:
                none_count += 1
        if none_count == len(info_file_sizes_lst):
            return pd.DataFrame()
        else:
            info_file_sizes_df = pd.concat(info_file_sizes_lst, ignore_index=True)
            return info_file_sizes_df

    def aggregate_model_failures(self) -> pd.DataFrame:
        model_failures_list = self.load_model_failures()
        none_count = 0
        for e in model_failures_list:
            if e is None:
                none_count += 1
        if none_count == len(model_failures_list):
            return pd.DataFrame()
        else:
            model_failures_df = pd.concat(model_failures_list, ignore_index=True)
            return model_failures_df

    def get_logs(self) -> List[str]:
        return self._loop_func(func=OutputContext.get_logs, input_list=self.output_contexts)

    def aggregate_logs(self) -> str:
        log_list = self.get_logs()
        return "\n\n".join(log_list)

    def get_amlb_info(self) -> List[str]:
        return self._loop_func(func=OutputContext.get_amlb_info, input_list=self.output_contexts)

    def get_benchmark_failures(self):
        amlb_info_dict = dict()

        amlb_info_list = self.get_amlb_info()

        for output_context, amlb_info in zip(self.output_contexts, amlb_info_list):
            path_relative = remove_prefix(output_context.path, prefix=self.path)

            if amlb_info is not None:
                if amlb_info in amlb_info_dict:
                    amlb_info_dict[amlb_info].append(path_relative)
                else:
                    amlb_info_dict[amlb_info] = [path_relative]
        amlb_info_count_dict = dict()
        for info, paths in amlb_info_dict.items():
            amlb_info_count_dict[info] = len(paths)
        sorted_info = sorted(amlb_info_count_dict.items(), key=lambda x: x[1], reverse=True)
        sorted_info = [i[0] for i in sorted_info]
        for i in sorted_info:
            count = amlb_info_count_dict[i]
            print(f"{count} | {i}\n" f"\t{amlb_info_dict[i]}")

    @staticmethod
    def _aggregate_leaderboards_ray(output_contexts, columns_to_keep, with_infer_speed):
        num_contexts = len(output_contexts)
        if not ray.is_initialized():
            ray.init()
        results = []
        for i, output_context in enumerate(output_contexts):
            results.append(
                get_single_leaderboard_ray.remote(output_context, columns_to_keep, with_infer_speed, i, num_contexts)
            )
        result = ray.get(results)
        result = [r for r in result if r is not None]
        return result

    @staticmethod
    def _aggregate_leaderboards_seq(output_contexts, columns_to_keep, with_infer_speed):
        num_contexts = len(output_contexts)
        results = []
        for i, output_context in enumerate(output_contexts):
            results.append(
                get_single_leaderboard_seq(output_context, columns_to_keep, with_infer_speed, i, num_contexts)
            )
        results = [r for r in results if r is not None]
        return results

    def construct_zs_dict(self, results_lst=None, zeroshot_metadata_list=None):
        if results_lst is None:
            results_lst = self.load_results()
        if zeroshot_metadata_list is None:
            zeroshot_metadata_list = self.load_zeroshot_metadata()
        return self._construct_zs_dict(results_lst=results_lst, zeroshot_metadata_list=zeroshot_metadata_list)

    @classmethod
    def _construct_zs_dict(cls, results_lst, zeroshot_metadata_list):
        num_paths = len(results_lst)
        aggregated_pred_proba = {}
        aggregated_ground_truth = {}
        for i, (zeroshot_metadata, scores) in enumerate(zip(zeroshot_metadata_list, results_lst)):
            id = scores["id"][0]
            fold = scores.iloc[0]["fold"]
            task_name = scores.iloc[0]["task"]
            fold = int(fold)
            print(f"{i + 1}/{num_paths} | {task_name} | {fold}")

            if task_name not in aggregated_ground_truth:
                aggregated_ground_truth[task_name] = {}
            if fold not in aggregated_ground_truth[task_name]:
                aggregated_ground_truth[task_name][fold] = {}
                for k in [
                    "y_val",
                    "y_test",
                    "eval_metric",
                    "problem_type",
                    "ordered_class_labels",
                    "ordered_class_labels_transformed",
                    "problem_type_transform",
                    "num_classes",
                    "label",
                ]:
                    aggregated_ground_truth[task_name][fold][k] = zeroshot_metadata[k]
                aggregated_ground_truth[task_name][fold]["task"] = task_name
                aggregated_ground_truth[task_name][fold]["id"] = id
            if task_name not in aggregated_pred_proba:
                aggregated_pred_proba[task_name] = {}
            if fold not in aggregated_pred_proba[task_name]:
                aggregated_pred_proba[task_name][fold] = {}
            for k in ["pred_proba_dict_val", "pred_proba_dict_test"]:
                if k not in aggregated_pred_proba[task_name][fold]:
                    aggregated_pred_proba[task_name][fold][k] = {}
                for m, pred_proba in zeroshot_metadata[k].items():
                    if aggregated_ground_truth[task_name][fold]["problem_type"] == "binary":
                        if isinstance(pred_proba, pd.DataFrame):
                            assert len(pred_proba.columns) == 2
                            pred_proba = pred_proba[1]
                        assert isinstance(pred_proba, pd.Series)
                    aggregated_pred_proba[task_name][fold][k][m] = pred_proba
        return aggregated_pred_proba, aggregated_ground_truth


def _with_seq(
    func, input_list: list, kwargs=None, allow_exception=False, exception_default=None, num_cpus=None
) -> list:
    """
    For-loop through a function call sequentially
    """
    if kwargs is None:
        kwargs = dict()
    if allow_exception:

        def _func(*args, **kw):
            try:
                return func(*args, **kw)
            except:
                return exception_default

    else:
        _func = func
    out_list = []
    for input_val in input_list:
        out_list.append(_func(input_val, **kwargs))
    return out_list


def _with_ray(
    func,
    input_list: list,
    kwargs=None,
    allow_exception=False,
    exception_default=None,
    num_cpus: int = None,
    progress_bar: bool = True,
) -> list:
    """
    For-loop through a function call with ray
    """
    if kwargs is None:
        kwargs = dict()
    if allow_exception:

        def _func(*args, **kw):
            try:
                return func(*args, **kw)
            except:
                return exception_default

    else:
        _func = func

    if not ray.is_initialized():
        num_cpus_system = os.cpu_count()
        if num_cpus is not None and num_cpus_system > num_cpus:
            # Restrict to avoid strange errors, probably S3 throttling
            ray.init(num_cpus=num_cpus)
        else:
            ray.init()
    remote_func = ray.remote(_func)
    results = []
    for i in input_list:
        results.append(remote_func.remote(i, **kwargs))
    if progress_bar:
        result = [ray.get(res) for res in tqdm(results)]
    else:
        result = ray.get(results)
    return result


@ray.remote
def get_single_leaderboard_ray(output_context: OutputContext, columns_to_keep, with_infer_speed, i, num_contexts):
    return output_context.get_single_leaderboard(columns_to_keep, with_infer_speed, i, num_contexts=num_contexts)


def get_single_leaderboard_seq(output_context: OutputContext, columns_to_keep, with_infer_speed, i, num_contexts):
    return output_context.get_single_leaderboard(columns_to_keep, with_infer_speed, i, num_contexts=num_contexts)


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    else:
        raise AssertionError("Lacking prefix!")



_TEXT = "text"
_NER = "ner"
_CATEGORICAL = "categorical"
_NUMERICAL = "numerical"
_BINARY = "binary"
_MULTICLASS = "multiclass"
_REGRESSION = "regression"
_IMAGE_SIMILARITY = "image_similarity"
_IMAGE_TEXT_SIMILARITY = "image_text_similarity"
_TEXT_SIMILARITY = "text_similarity"
_OBJECT_DETECTION = "object_detection"
_FEW_SHOT_CLASSIFICATION = "few_shot_classification"



from __future__ import annotations

import re

import numpy as np
import pandas as pd

from ..constants import *
from . import preprocess_utils


def _update_framework_name(name: str, parent: str, name_suffix_1: str, name_suffix_2: str):
    is_valid_parent = parent and isinstance(parent, str)
    if is_valid_parent:
        name_new = parent
    else:
        name_new = name
    if name_suffix_1:
        name_new = name_new + "_" + str(name_suffix_1)
    if name_suffix_2:
        name_new = name_new + "_" + str(name_suffix_2)
    if is_valid_parent:
        return name, name_new
    else:
        return name_new, parent


def _parent_rename(parent: str | None, name: str):
    if parent and isinstance(parent, str):
        rename = parent + "_" + name
    else:
        rename = name
    return rename


def preprocess_openml_input(
    df: pd.DataFrame,
    framework_suffix=None,
    framework_rename_dict=None,
    folds_to_keep=None,
    framework_suffix_column=None,
):
    raw_input = df.copy()
    assert len(raw_input.columns) == len(set(raw_input.columns))
    raw_input = _rename_openml_columns(raw_input)
    raw_input = raw_input[raw_input[FRAMEWORK].notnull()]
    if framework_rename_dict is not None:
        for key in framework_rename_dict.keys():
            raw_input[FRAMEWORK] = [
                framework_rename_dict[key] if framework[0] == key else framework[0]
                for framework in zip(raw_input[FRAMEWORK])
            ]

    if framework_suffix_column:
        framework_suffix_column_to_zip = [f for f in raw_input[framework_suffix_column]]
    else:
        framework_suffix_column_to_zip = [None] * len(raw_input)
    if "framework_parent" in raw_input.columns:
        framework_parent_to_zip = [f for f in raw_input["framework_parent"]]
    else:
        framework_parent_to_zip = [None] * len(raw_input)

    updated_names_and_parent = [
        _update_framework_name(
            name=name,
            parent=parent,
            name_suffix_1=name_suffix_1,
            name_suffix_2=framework_suffix,
        )
        for (name, parent, name_suffix_1) in zip(
            raw_input[FRAMEWORK], framework_parent_to_zip, framework_suffix_column_to_zip
        )
    ]

    updated_names = [name for name, parent in updated_names_and_parent]
    updated_parents = [parent for name, parent in updated_names_and_parent]

    raw_input[FRAMEWORK] = updated_names
    if "framework_parent" in raw_input.columns:
        raw_input["framework_parent"] = updated_parents

    if "framework_parent" in raw_input:
        # raw_input["framework_child"] = raw_input[FRAMEWORK]
        raw_input["framework_child"] = np.where(raw_input["framework_parent"].isnull(), np.nan, raw_input[FRAMEWORK])

    if "framework_parent" in raw_input:
        framework_parent_to_zip = [f for f in raw_input["framework_parent"]]
    else:
        framework_parent_to_zip = [None] * len(raw_input)

    raw_input[FRAMEWORK] = [
        _parent_rename(
            parent=parent,
            name=name,
        )
        for (parent, name) in zip(framework_parent_to_zip, raw_input[FRAMEWORK])
    ]

    # TODO: This is a hack and won't work for all metrics, metric_error should ideally be calculated prior to preprocessing
    metric_list = [
        "auc",
        "acc",
        "accuracy",
        "balacc",
        "map",
        "roc_auc",
        "r2",
        "coverage",
        "f1",
        "f1_macro",
        "f1_micro",
        "quadratic_kappa",
    ]
    if METRIC_ERROR not in raw_input:
        raw_input[METRIC_ERROR] = [
            1 - score if metric in metric_list else float(score) if metric == "rmse" else -score
            for score, metric in zip(raw_input[METRIC_SCORE], raw_input["metric"])
        ]

    if raw_input[METRIC_ERROR].min() < 0:
        eps = -1 / 1e8
        num_negative = len(raw_input[raw_input[METRIC_ERROR] < 0])

        if raw_input[METRIC_ERROR].min() < eps:
            raise AssertionError(
                f"METRIC_ERROR cannot be negative! There may be a bug. "
                f"Found min value: {raw_input[METRIC_ERROR].min()}. "
                f"{num_negative} rows had negative values!"
            )
        else:
            print(
                f"WARNING: min METRIC_ERROR was found to be negative, but was higher than epsilon {eps}! "
                f"({raw_input[METRIC_ERROR].min()}) {num_negative} rows had negative values! "
                f"Setting all negative values to 0."
            )
            raw_input.loc[raw_input[METRIC_ERROR] < 0, METRIC_ERROR] = 0
    assert raw_input[METRIC_ERROR].min() >= 0

    cleaned_input = preprocess_utils.clean_result(raw_input, folds_to_keep=folds_to_keep, remove_invalid=False)

    if "tid" not in cleaned_input:
        if "id" in cleaned_input:
            cleaned_input["tid"] = [
                int(part) for x in cleaned_input["id"] for part in re.split(r"[/_]", x)[-1:] if part.isdigit()
            ]

    if "tid" in cleaned_input:
        """
        Update tid to the new ones in case the runs are old

        Name                | oldtid | newtid

        KDDCup09-Upselling  | 360115 | 360975
        MIP-2016-regression | 359947 | 360945
        QSAR-TID-10980      |  14097 | 360933
        QSAR-TID-11         |  13854 | 360932
        """
        cleaned_input["tid"] = (
            cleaned_input["tid"]
            .map(
                {
                    360115: 360975,
                    359947: 360945,
                    14097: 360933,
                    13854: 360932,
                }
            )
            .fillna(cleaned_input["tid"])
        )

    required_output_columns = [
        DATASET,
        FOLD,
        FRAMEWORK,
        METRIC_ERROR,
        METRIC,
        PROBLEM_TYPE,
        TIME_TRAIN_S,
        TIME_INFER_S,
    ]
    actual_columns = list(cleaned_input.columns)
    missing_columns = [c for c in required_output_columns if c not in actual_columns]
    if missing_columns:
        raise AssertionError(f"Missing expected output columns: {missing_columns}")

    reordered_columns = required_output_columns + [c for c in actual_columns if c not in required_output_columns]
    cleaned_input = cleaned_input[reordered_columns]

    return cleaned_input


def _update_and_drop_column(df: pd.DataFrame, col_to_update: str, col_to_drop: str) -> pd.DataFrame:
    if col_to_drop in df:
        if col_to_update in df:
            df[col_to_update] = df[col_to_drop].fillna(df[col_to_update])
            df = df.drop(columns=[col_to_drop])
        else:
            df = df.rename(columns={col_to_drop: col_to_update})
    return df


def _rename_openml_columns(result_df):
    rename_order = [
        ("type", PROBLEM_TYPE),
        (TASK, DATASET),
        (RESULT, METRIC_SCORE),
        (DURATION, TIME_TRAIN_S),
        (PREDICT_DURATION, TIME_INFER_S),
        ("training_duration", TIME_TRAIN_S),
        ("score_test", METRIC_SCORE),
        ("pred_time_test", TIME_INFER_S),
    ]
    renamed_df = result_df.copy()
    for col_to_drop, col_to_update in rename_order:
        renamed_df = _update_and_drop_column(renamed_df, col_to_update=col_to_update, col_to_drop=col_to_drop)
    return renamed_df



import logging
import os

import pandas as pd
import yaml

from autogluon.bench.utils.dataset_utils import get_data_home_dir
from autogluon.common.loaders._utils import download

logger = logging.getLogger(__name__)


class TextDataLoader:
    def __init__(
        self,
        dataset_name: str,
        dataset_config_file: str,
        split: str = "train",
        lang: str = "en",
        fewshot: bool = False,
        shot: int = 50,
        seed: int = 0,
    ):
        with open(dataset_config_file, "r") as f:
            config = yaml.safe_load(f)
        self.dataset_config = config[dataset_name]
        if split == "val":
            split = "validation"
        if split not in self.dataset_config["splits"]:
            logger.warning(f"Data split {split} not available.")
            self.data = None
            return

        self.name = dataset_name
        self.split = split

        subsample_path = self.dataset_config["subsample_path"].format(shot=shot, seed=seed)
        url = self.dataset_config["url"].format(
            name=self.name,
            lang=lang,
            subsample_path=subsample_path if fewshot and self.split in self.dataset_config["subsample_splits"] else "",
            split=self.split,
        )
        base_dir = get_data_home_dir()
        data_dir = os.path.join(self.name, lang)
        if fewshot:
            data_dir = os.path.join(data_dir, "subsampling", f"{shot}_shot-seed{seed}")
        self.dataset_dir = os.path.join(base_dir, data_dir)
        data_path = os.path.join(self.dataset_dir, f"{split}.csv")
        download(url, path=data_path)

        self.data = pd.read_csv(
            data_path,
            header=None,
            names=self.dataset_config["data_columns"],
            sep=self.dataset_config.get("data_sep", ","),
            on_bad_lines="warn",
        )

    @property
    def problem_type(self):
        return self.dataset_config["problem_type"]

    @property
    def metric(self):
        return self.dataset_config["metric"]

    @property
    def feature_columns(self):
        return self.dataset_config["feature_columns"]

    @property
    def label_columns(self):
        return self.dataset_config["label_columns"]



import importlib.resources
import json
import os
import shutil
import subprocess
import tempfile
from typing import Optional

import boto3
import typer
import yaml

with importlib.resources.path("autogluon.bench.cloud.aws", "stack_handler.py") as file_path:
    module_base_dir = os.path.dirname(file_path)
CONTEXT_FILE = "./cdk.context.json"

app = typer.Typer()


def get_instance_type_specs(instance_type, region):
    ec2_client = boto3.client("ec2", region_name=region)
    response = ec2_client.describe_instance_types(InstanceTypes=[instance_type])

    instance_type_info = response["InstanceTypes"][0]

    gpu_info_list = instance_type_info.get("GpuInfo", {}).get("Gpus", [{}])
    gpu_count = sum(gpu_info.get("Count", 0) for gpu_info in gpu_info_list)

    vcpu_info = instance_type_info.get("VCpuInfo", {})
    vcpu_count = vcpu_info.get("DefaultVCpus", 0)

    memory_info = instance_type_info.get("MemoryInfo", {})
    memory = memory_info.get("SizeInMiB", 0)

    return gpu_count, vcpu_count, memory


def _get_temp_cdk_app_path():
    temp_dir = tempfile.mkdtemp()
    temp_cdk_app_path = os.path.join(temp_dir, "app.py")
    with importlib.resources.path("autogluon.bench.cloud.aws", "app.py") as cdk_app_path:
        shutil.copy2(cdk_app_path, temp_cdk_app_path)
    os.chmod(temp_cdk_app_path, 0o755)
    return temp_cdk_app_path


def construct_context(custom_configs: dict) -> dict:
    """
    Constructs the AWS Cloud Development Kit (CDK) context using a combination of default configuration
    settings and custom settings, and writes the context to a JSON file. Also sets environment variables for
    the CDK deployment account and region.

    Args:
        custom_configs (dict, optional): A dictionary containing custom configuration settings. Defaults to {}.

    Returns:
        dict: A dictionary containing the constructed CDK context settings.
    """
    default_config_file = os.path.join(module_base_dir, "default_config.yaml")
    configs = {}
    with open(default_config_file, "r") as f:
        configs = yaml.safe_load(f)
    configs.update(custom_configs)
    prefix = configs["PREFIX"]
    gpu_count, vcpu_count, memory = get_instance_type_specs(
        instance_type=configs["INSTANCE"], region=configs["CDK_DEPLOY_REGION"]
    )
    configs["TIME_LIMIT"] = (
        configs["TIME_LIMIT"] + 3600
    )  # buffer to account for instance start, dataset download and overhead

    context_to_parse = {
        "CDK_DEPLOY_ACCOUNT": configs["CDK_DEPLOY_ACCOUNT"],
        "CDK_DEPLOY_REGION": configs["CDK_DEPLOY_REGION"],
        "STACK_NAME_PREFIX": prefix,  # aws resource tag key, also used as name prefix for resources created
        "STACK_NAME_TAG": "benchmark",  # aws resource tag value
        "STATIC_RESOURCE_STACK_NAME": f"{prefix}-static-resource-stack",
        "BATCH_STACK_NAME": f"{prefix}-batch-stack",
        "METRICS_BUCKET": configs["METRICS_BUCKET"],  # bucket to upload metrics
        "DATA_BUCKET": configs.get("DATA_BUCKET", None),  # bucket to download data
        "INSTANCE_TYPES": [configs["INSTANCE"]],  # can be a list of instance families or instance types
        "COMPUTE_ENV_MAXV_CPUS": vcpu_count
        * configs["MAX_MACHINE_NUM"],  # total max v_cpus in batch compute environment
        "CONTAINER_GPU": gpu_count,  # GPU reserved for container
        "CONTAINER_VCPU": vcpu_count,  # v_cpus reserved for container
        "CONTAINER_MEMORY": memory
        - configs[
            "RESERVED_MEMORY_SIZE"
        ],  # memory in MB reserved for container, also used for shm_size, i.e. `shared_memory_size`
        "TIME_LIMIT": configs["TIME_LIMIT"],  # timeout of instance, in seconds
        "BLOCK_DEVICE_VOLUME": configs["BLOCK_DEVICE_VOLUME"],  # device attached to instance, in GB
        "LAMBDA_FUNCTION_NAME": f"{prefix}-batch-job-function",
        "VPC_NAME": configs.get(
            "VPC_NAME", None
        ),  # it's recommended to share a vpc for all benchmark infra, you can lookup an existing VPC name under aws console -> VPC, if you want to create a new one, assign a new name
    }
    with open(CONTEXT_FILE, "w+") as f:
        try:
            cdk_config = json.load(f)
        except:
            cdk_config = {}
        cdk_config.update(context_to_parse)
        json.dump(cdk_config, f, indent=2)
        f.close()
    # set environment variables
    os.environ["CDK_DEPLOY_ACCOUNT"] = str(configs["CDK_DEPLOY_ACCOUNT"])
    os.environ["CDK_DEPLOY_REGION"] = configs["CDK_DEPLOY_REGION"]

    return context_to_parse


def deploy_stack(custom_configs: dict) -> dict:
    """
    Deploys the AWS CloudFormation stack containing the benchmarking infrastructure by calling the deploy.sh
    script and passing it the required command line arguments. Constructs the CDK context using the custom
    configuration settings specified in the configs parameter, or the default configuration settings if no
    custom settings are provided.

    Args:
        configs (dict, optional): A dictionary containing custom configuration settings. Defaults to None.

    Returns:
        dict: A dictionary containing the CDK context settings used for the deployment.
    """
    cdk_path = _get_temp_cdk_app_path()
    custom_infra_configs = custom_configs.get("cdk_context", {})
    infra_configs = construct_context(custom_configs=custom_infra_configs)
    instance_type: str = infra_configs["INSTANCE_TYPES"][0]
    os.environ["AG_BENCH_BASE_IMAGE"] = "nvidia/cuda:12.2.0-runtime-ubuntu20.04"
    if not instance_type.startswith(("p", "g")):
        # CPU instances
        os.environ["AG_BENCH_BASE_IMAGE"] = "ubuntu:20.04"
    command = [
        os.path.join(module_base_dir, "deploy.sh"),
        infra_configs["STATIC_RESOURCE_STACK_NAME"],
        infra_configs["BATCH_STACK_NAME"],
        cdk_path,
    ]

    subprocess.check_call(command)
    shutil.rmtree(os.path.dirname(cdk_path))

    return infra_configs


@app.command()
def destroy_stack(
    static_resource_stack: Optional[str] = typer.Option(None, help="The static resource stack name."),
    batch_stack: Optional[str] = typer.Option(None, help="The batch stack name."),
    cdk_deploy_account: Optional[str] = typer.Option(None, help="The CDK deploy account ID."),
    cdk_deploy_region: Optional[str] = typer.Option(None, help="The CDK deploy region."),
    config_file: Optional[str] = typer.Option(None, help="Path to YAML config file containing stack information."),
):
    """
    This function destroys AWS CloudFormation stacks using the AWS Cloud Development Kit (CDK).

    It first sets up the necessary environment variables for the CDK, then calls a shell script
    that uses the CDK to destroy the specified static resource stack and batch stack. Finally, it
    removes the temporary directory that was used to deploy the CDK app.

    If you have previously deployed with `agbench run CONFIG_FILE,`
    you can find the AWS configs saved under {root_dir}/{module}/{prefix}_{timestamp}/aws_configs.yaml"
    """
    cdk_path = _get_temp_cdk_app_path()

    if config_file is not None:
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)
            static_resource_stack = config.get("STATIC_RESOURCE_STACK_NAME", static_resource_stack)
            batch_stack = config.get("BATCH_STACK_NAME", batch_stack)
            cdk_deploy_account = config.get("CDK_DEPLOY_ACCOUNT", cdk_deploy_account)
            cdk_deploy_region = config.get("CDK_DEPLOY_REGION", cdk_deploy_region)

    if static_resource_stack is None or batch_stack is None or cdk_deploy_account is None or cdk_deploy_region is None:
        raise ValueError(
            "static_resource_stack, batch_stack, cdk_deploy_account and cdk_deploy_region must be specified or configured in the config_file."
        )

    os.environ["CDK_DEPLOY_ACCOUNT"] = cdk_deploy_account
    os.environ["CDK_DEPLOY_REGION"] = cdk_deploy_region
    subprocess.check_call(
        [
            os.path.join(module_base_dir, "destroy.sh"),
            static_resource_stack,
            batch_stack,
            cdk_deploy_region,
            cdk_path,
        ]
    )
    shutil.rmtree(os.path.dirname(cdk_path))



import copy
import math

import numpy as np
import pandas as pd

from .constants import *


def compare_frameworks(
    results_raw,
    frameworks=None,
    banned_datasets=None,
    folds_to_keep=None,
    filter_errors=True,
    verbose=True,
    columns_to_agg_extra=None,
    datasets=None,
):
    columns_to_agg = [DATASET, FRAMEWORK, PROBLEM_TYPE, TIME_TRAIN_S, METRIC_ERROR]
    if columns_to_agg_extra:
        columns_to_agg += columns_to_agg_extra
    if frameworks is None:
        frameworks = sorted(list(results_raw[FRAMEWORK].unique()))

    if filter_errors:  # FIXME: This should not be toggled, instead filter_errors should be passed to filter_results
        results = filter_results(
            results_raw=results_raw,
            valid_frameworks=frameworks,
            banned_datasets=banned_datasets,
            folds_to_keep=folds_to_keep,
        )
        if len(results) == 0:
            frameworks_to_try = []
            unique_counts = results_raw.value_counts(FRAMEWORK)
            valid_frameworks_to_try = [f for f in unique_counts.index if f in frameworks]
            print(
                f"Error! No results after filtering! "
                f"Iteratively adding frameworks starting from most successes to least successes to identify problematic frameworks.\n"
                f"+ (Newest Added Framework): (Number of successful tasks)"
            )
            max_successful_frameworks = []
            for f in valid_frameworks_to_try:
                frameworks_to_try.append(f)
                results_test = filter_results(
                    results_raw=results_raw,
                    valid_frameworks=frameworks_to_try,
                    banned_datasets=banned_datasets,
                    folds_to_keep=folds_to_keep,
                )
                num_successes = len(results_test.value_counts([DATASET, FOLD]))
                if num_successes == 0:
                    print_suffix = " (No successes! Exception will be raised.)"
                else:
                    print_suffix = ""
                    max_successful_frameworks = copy.deepcopy(frameworks_to_try)
                print(f"+ {f}: {num_successes}{print_suffix}")
            print(f"Valid Frameworks Prior to Reaching 0 Successes:\n" f"\t{max_successful_frameworks}")
            raise AssertionError(f"No results exist after filtering errors!\n" f"Unique Counts:\n{unique_counts}")
    else:
        results = results_raw.copy()

    results_agg = results[columns_to_agg].groupby([DATASET, FRAMEWORK, PROBLEM_TYPE]).mean().reset_index()

    worst_scores = results_agg.sort_values(METRIC_ERROR, ascending=False).drop_duplicates(DATASET)
    worst_scores = worst_scores[[DATASET, METRIC_ERROR]]
    worst_scores.columns = [DATASET, "WORST_ERROR"]
    best_scores = results_agg.sort_values(METRIC_ERROR, ascending=True).drop_duplicates(DATASET)
    best_scores = best_scores[[DATASET, METRIC_ERROR]]
    best_scores.columns = [DATASET, "BEST_ERROR"]

    results_agg = results_agg.merge(best_scores, on=DATASET)
    results_agg = results_agg.merge(worst_scores, on=DATASET)
    results_agg[BESTDIFF] = 1 - (results_agg["BEST_ERROR"] / results_agg[METRIC_ERROR])
    results_agg[LOSS_RESCALED] = (results_agg[METRIC_ERROR] - results_agg["BEST_ERROR"]) / (
        results_agg["WORST_ERROR"] - results_agg["BEST_ERROR"]
    )
    results_agg[BESTDIFF] = results_agg[BESTDIFF].fillna(0)
    results_agg[LOSS_RESCALED] = results_agg[LOSS_RESCALED].fillna(0)
    results_agg = results_agg.drop(["BEST_ERROR"], axis=1)
    results_agg = results_agg.drop(["WORST_ERROR"], axis=1)

    for time_attr in [TIME_TRAIN_S, TIME_INFER_S]:
        if time_attr in columns_to_agg:
            best_time_attr = "BEST_" + time_attr
            best_speed = (
                results_agg[[DATASET, time_attr]].sort_values(time_attr, ascending=True).drop_duplicates(DATASET)
            )
            best_speed.columns = [DATASET, best_time_attr]
            results_agg = results_agg.merge(best_speed, on=DATASET)
            results_agg[time_attr + "_rescaled"] = results_agg[time_attr] / results_agg[best_time_attr]
            results_agg = results_agg.drop([best_time_attr], axis=1)

    valid_tasks = list(results_agg[DATASET].unique())

    results_ranked, results_ranked_by_dataset = rank_result(results_agg)
    rank_1 = results_ranked_by_dataset[results_ranked_by_dataset[RANK] == 1]
    rank_1_count = rank_1[FRAMEWORK].value_counts()
    results_ranked["rank=1_count"] = rank_1_count
    results_ranked["rank=1_count"] = results_ranked["rank=1_count"].fillna(0).astype(int)

    rank_2 = results_ranked_by_dataset[(results_ranked_by_dataset[RANK] > 1) & (results_ranked_by_dataset[RANK] <= 2)]
    rank_2_count = rank_2[FRAMEWORK].value_counts()

    results_ranked["rank=2_count"] = rank_2_count
    results_ranked["rank=2_count"] = results_ranked["rank=2_count"].fillna(0).astype(int)

    rank_3 = results_ranked_by_dataset[(results_ranked_by_dataset[RANK] > 2) & (results_ranked_by_dataset[RANK] <= 3)]
    rank_3_count = rank_3[FRAMEWORK].value_counts()

    results_ranked["rank=3_count"] = rank_3_count
    results_ranked["rank=3_count"] = results_ranked["rank=3_count"].fillna(0).astype(int)

    rank_l3 = results_ranked_by_dataset[(results_ranked_by_dataset[RANK] > 3)]
    rank_l3_count = rank_l3[FRAMEWORK].value_counts()

    results_ranked["rank>3_count"] = rank_l3_count
    results_ranked["rank>3_count"] = results_ranked["rank>3_count"].fillna(0).astype(int)

    if datasets is None:
        datasets = sorted(list(results_ranked_by_dataset[DATASET].unique()))
    datasets_len = len(datasets)
    errors_list = []
    for framework in frameworks:
        results_framework = filter_results(
            results_raw=results_raw,
            valid_frameworks=[framework],
            banned_datasets=banned_datasets,
            folds_to_keep=folds_to_keep,
        )
        results_framework_agg = (
            results_framework[columns_to_agg].groupby([DATASET, FRAMEWORK, PROBLEM_TYPE]).mean().reset_index()
        )
        num_valid = len(results_framework_agg[results_framework_agg[FRAMEWORK] == framework])
        num_errors = datasets_len - num_valid
        errors_list.append(num_errors)
    errors_series = pd.Series(data=errors_list, index=frameworks)

    results_ranked["error_count"] = errors_series
    results_ranked["error_count"] = results_ranked["error_count"].fillna(0).astype(int)
    results_ranked = results_ranked.reset_index()

    if verbose:
        print("valid_tasks:", len(valid_tasks))
        with pd.option_context("display.max_rows", None, "display.max_columns", None, "display.width", 1000):
            print(results_ranked)
        print()
    return results_ranked, results_ranked_by_dataset


def filter_results(results_raw, valid_frameworks, banned_datasets=None, folds_to_keep=None):
    results = results_raw.copy()
    if folds_to_keep is not None:
        results = results[results[FOLD].isin(folds_to_keep)]
    if banned_datasets is not None:
        results = results[~results[DATASET].isin(banned_datasets)]
    results = keep_only_valid_datasets(results, valid_models=valid_frameworks)

    return results


def keep_only_valid_datasets(result_df, valid_models):
    result_df = result_df[result_df[FRAMEWORK].isin(valid_models)]
    tasks = list(result_df[DATASET].unique())
    frameworks = list(result_df[FRAMEWORK].unique())

    fw_datasets = {}
    for framework in frameworks:
        fw_datasets[framework] = set(list(result_df[result_df[FRAMEWORK] == framework][DATASET].unique()))
    valid_tasks = []
    for task in tasks:
        skip_task = False
        for framework in frameworks:
            if task not in fw_datasets[framework]:
                skip_task = True
                break
        if not skip_task:
            valid_tasks.append(task)
    if len(valid_tasks) == 0:
        valid_result_df = pd.DataFrame(columns=result_df.columns)
    else:
        valid_tasks_set = set(valid_tasks)
        valid_result_df = result_df[result_df[DATASET].isin(valid_tasks_set)].reset_index(drop=True)

    return valid_result_df


def rank_result(result_df):
    datasets = list(result_df[DATASET].unique())
    result_df = result_df.copy()
    result_df[METRIC_ERROR] = [round(x[0], 5) for x in zip(result_df[METRIC_ERROR])]
    num_frameworks = len(result_df[FRAMEWORK].unique())
    dfs = []
    if num_frameworks == 1:
        sorted_df_full = result_df
        sorted_df_full[RANK] = 1
    else:
        for dataset in datasets:
            dataset_df = result_df[result_df[DATASET] == dataset]
            sorted_df = dataset_df.copy()
            sorted_df[RANK] = sorted_df[METRIC_ERROR].rank()
            dfs.append(sorted_df)
        sorted_df_full = pd.concat(dfs, ignore_index=True)
    model_ranks_df = sorted_df_full.groupby([FRAMEWORK]).mean(numeric_only=True).sort_values(by=RANK)
    return model_ranks_df, sorted_df_full


def graph_vs(results_df: pd.DataFrame, f1: str, f2: str, z_stats: pd.Series = None, datasets: list = None):
    results_df = results_df[results_df[FRAMEWORK].isin([f1, f2])][[FRAMEWORK, DATASET, PROBLEM_TYPE, METRIC_ERROR]]
    results_df = results_df.groupby([FRAMEWORK, DATASET, PROBLEM_TYPE]).mean(numeric_only=True).reset_index()
    problem_type_dict = results_df.set_index(DATASET)[PROBLEM_TYPE].to_dict()
    results_1 = results_df[results_df[FRAMEWORK] == f1]
    results_1 = results_1.set_index(DATASET)[METRIC_ERROR].to_dict()
    results_2 = results_df[results_df[FRAMEWORK] == f2]
    results_2 = results_2.set_index(DATASET)[METRIC_ERROR].to_dict()

    if datasets is None:
        datasets = list(set(results_1.keys()).intersection(results_2.keys()))
        if z_stats is not None:  # FIXME: need to filter missing folds still
            datasets = list(set(datasets).intersection(set(z_stats.index)))

    import matplotlib.pyplot as plt
    from matplotlib import colors

    fig, ax = plt.subplots(dpi=300)  # Create a figure containing a single axes.
    ax.set(xlabel=f2, ylabel=f1, title="Dataset Error Rate Comparison")

    results_1_lst = []
    results_2_lst = []
    r_1_dict = dict()
    r_2_dict = dict()
    z_stat_lst = []

    win_1 = 0
    win_2 = 0
    tie = 0

    from collections import defaultdict

    problem_type_d = defaultdict(list)
    problem_type_lst = []
    for dataset in datasets:
        r_1 = results_1[dataset]
        r_2 = results_2[dataset]

        problem_type = problem_type_dict[dataset]
        problem_type_d[problem_type].append(dataset)

        if r_1 == r_2 or abs(z_stats[dataset]) < 2:
            tie += 1
        elif r_1 > r_2:
            win_2 += 1
        else:
            win_1 += 1
        z_stat_lst.append(-z_stats[dataset])
        max_r = max(r_1, r_2)
        if max_r > 1:
            r_1 = r_1 / max_r
            r_2 = r_2 / max_r

        results_1_lst.append(r_1)
        results_2_lst.append(r_2)
        r_1_dict[dataset] = r_1
        r_2_dict[dataset] = r_2

        # problem_type_lst.append(marker)
        # if z_stats is not None:
        #     ax.scatter(r_1, r_2, c=z_stats[dataset], label=dataset)
        # else:
        #     ax.scatter(r_1, r_2, label=dataset)
    ax.plot((0, 1))
    cm = plt.cm.get_cmap("RdYlBu")

    for pt in problem_type_d:
        if pt == "binary":
            marker = "o"
        elif pt == "multiclass":
            marker = "^"
        elif pt == "regression":
            marker = "*"
        elif pt == "timeseries":
            marker = "P"
        else:
            raise AssertionError(f"Unknown problem type: {pt}")

        pt_datasets = problem_type_d[pt]
        r_1_lst = [r_1_dict[d] for d in pt_datasets]
        r_2_lst = [r_2_dict[d] for d in pt_datasets]
        z_stats_lst = [-z_stats[d] for d in pt_datasets]
        im = ax.scatter(
            r_2_lst,
            r_1_lst,
            c=z_stats_lst,
            cmap=cm,
            norm=colors.Normalize(vmin=-3, vmax=3, clip=True),
            edgecolor="k",
            linewidths=0.5,
            marker=marker,
            label=pt,
        )

        textstr = f"Win 1: {win_1}, Win 2: {win_2}, Tie: {tie}"
        props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)
        ax.text(0.05, 0.75, textstr, transform=ax.transAxes, fontsize=8, verticalalignment="top", bbox=props)

        ax.grid()
        # ax.set_xlim(0, 0.02)
        # ax.set_ylim(0, 0.02)
        # ax.set_xscale('log')
        # ax.set_yscale('log')
        fig.colorbar(im, ax=ax)
        plt.legend()
        plt.show()


# TODO: USE T-STATISTIC?
def compute_stderr_z_stat(
    results_df: pd.DataFrame, f1: str, f2: str, folds: list = None, win_z_threshold: float = 1.96, verbose: bool = True
) -> pd.Series:
    """
    Compute z-scores for each dataset by comparing f1 and f2.

    COMPARE BY:

    FOR EACH FOLD 1-10

    A1 - B1 = C1
    A2 - B2 = C2
    ...

    stddev of C1 - C10
    Calc probability Mean >0

    Parameters
    ----------
    results_df : :class:`pd.DataFrame`
        The results DataFrame. Must contain `dataset`, `fold`, `framework`, and `metric_error` columns.
    f1 : str
        The name of the first framework to compare.
    f2 : str
        The name of the second framework to compare.
    folds : list, default None
        The list of folds to use in the calculation.
        If None, infers based on folds used in f1 and f2.
    win_z_threshold : float, default 1.96
        The z-score threshold for declaring a framework as winning a dataset.
        If the absolute z-score is less than this value, then the dataset is considered a tie.
        Only relevant if `verbose=True`.
    verbose : bool, default True
        Whether to print info such as wins, ties, and a preview of the output pandas Series.

    Returns
    -------
    Returns a pandas Series where the index is the dataset and the value is the z-statistic. Positive z-statistics favor f1.

    """
    if f1 == f2:
        raise AssertionError("f1 and f2 cannot be the same!")

    results_df = results_df[[FRAMEWORK, DATASET, FOLD, METRIC_ERROR]]
    results_df = results_df[results_df[FRAMEWORK].isin([f1, f2])]
    if folds is None:
        folds = list(results_df[FOLD].unique())
    num_folds = len(folds)
    if num_folds <= 1:
        raise AssertionError("Not enough folds to calculate stderr")
    results_df = results_df[results_df[FOLD].isin(folds)]

    mean_std_error_df = results_df.groupby([FRAMEWORK, DATASET])[METRIC_ERROR].agg(["mean", "std", "count"])
    mean_std_error_df = mean_std_error_df.reset_index()
    dataset_row_count = mean_std_error_df.groupby(DATASET)["count"].sum()
    valid_datasets = list(dataset_row_count[dataset_row_count == num_folds * 2].index)

    results1 = results_df[results_df[FRAMEWORK] == f1].set_index([DATASET, FOLD])
    results2 = results_df[results_df[FRAMEWORK] == f2].set_index([DATASET, FOLD])

    results_diff = results2[METRIC_ERROR] - results1[METRIC_ERROR]

    datasets_z_statistic = {}
    for dataset in valid_datasets:
        diff_per_fold = results_diff[results_diff.index.get_level_values(DATASET) == dataset]
        mean_diff = diff_per_fold.mean()
        std_diff = diff_per_fold.std()
        stderr_diff = std_diff / math.sqrt(len(diff_per_fold))
        if stderr_diff == 0:
            if mean_diff > 0:
                z_diff = 99
            elif mean_diff < 0:
                z_diff = -99
            else:
                z_diff = 0
        else:
            z_diff = mean_diff / stderr_diff

        datasets_z_statistic[dataset] = z_diff

    win1 = 0
    win2 = 0
    tie = 0

    z_stat_series = pd.Series(datasets_z_statistic)
    z_stat_series = z_stat_series.sort_values(ascending=False)

    for dataset in datasets_z_statistic:
        z_stat = datasets_z_statistic[dataset]
        if z_stat > win_z_threshold:
            win1 += 1
        elif z_stat < -win_z_threshold:
            win2 += 1
        else:
            tie += 1
    if verbose:
        print(z_stat_series)
        print(f"Win {win1} : {f1}")
        print(f"Win {win2} : {f2}")
        print(f"Tie {tie}")

    return z_stat_series


# TODO: document
def compute_stderr_z_stat_bulk(
    framework: str,
    frameworks_to_compare: list,
    results_raw: pd.DataFrame,
    folds: list = None,
    verbose: bool = True,
) -> pd.DataFrame:
    """
    Compares framework with each framework in frameworks_to_compare, constructing an overall result DataFrame as output.

    :param framework:
    :param frameworks_to_compare:
    :param results_raw:
    :param folds:
    :param verbose:
    :return:
    """
    assert framework not in frameworks_to_compare
    z_stat_dict = {}
    for f2 in frameworks_to_compare:
        z_stat_series = compute_stderr_z_stat(results_raw, f1=framework, f2=f2, folds=folds, verbose=verbose)
        z_stat_dict[f2] = z_stat_series

    z_stat_df = pd.DataFrame(z_stat_dict)
    z_stat_df["min"] = z_stat_df.min(axis=1)
    z_stat_df = z_stat_df.sort_values(by=["min"], ascending=False)

    if verbose:
        print("Z-statistic comparisons:")
        with pd.option_context("display.max_columns", None, "display.width", 1000):
            print(z_stat_df)
    return z_stat_df


def compute_win_rate_per_dataset(
    f1: str,
    f2: str,
    results_raw: pd.DataFrame,
    folds: list = None,
    verbose: bool = True,
    epsilon=1e-5,
    epsilon_bestdiff=1e-4,
):
    if f1 == f2:
        raise AssertionError("f1 and f2 cannot be the same!")

    results_df = results_raw[[FRAMEWORK, DATASET, FOLD, METRIC_ERROR]]
    results_df = results_df[results_df[FRAMEWORK].isin([f1, f2])]
    if folds is None:
        folds = list(results_df[FOLD].unique())
    num_folds = len(folds)
    # if num_folds <= 1:
    #     raise AssertionError('Not enough folds to calculate stderr')
    results_df = results_df[results_df[FOLD].isin(folds)]

    dataset_row_count = results_df[DATASET].value_counts()
    valid_datasets = list(dataset_row_count.index)
    # valid_datasets = list(dataset_row_count[dataset_row_count == num_folds*2].index)

    out_dict = dict()

    for dataset in valid_datasets:
        bestdiff_dataset = []
        f1_wins = 0
        num_ties = 0
        valid_folds = 0
        dataset_df = results_df[results_df[DATASET] == dataset]
        for fold in folds:
            dataset_df_fold = dataset_df[dataset_df[FOLD] == fold]
            dataset_fold_dict = dataset_df_fold.set_index(FRAMEWORK)[METRIC_ERROR].to_dict()
            # print(dataset_fold_dict)
            if len(dataset_fold_dict.keys()) != 2:
                continue
            valid_folds += 1
            f1_metric_error = dataset_fold_dict[f1]
            f2_metric_error = dataset_fold_dict[f2]

            bestdiff = 0
            tie = False
            if f1_metric_error == f2_metric_error:
                tie = True
            elif epsilon is not None and abs(f1_metric_error - f2_metric_error) < epsilon:
                tie = True
                # print(f'within epsilon {dataset} {fold} | {f1_metric_error} | {f2_metric_error}')
            elif f1_metric_error < f2_metric_error:
                bestdiff = 1 - f1_metric_error / f2_metric_error
            else:
                bestdiff = -(1 - f2_metric_error / f1_metric_error)
            if not tie and epsilon_bestdiff is not None and abs(bestdiff) < epsilon_bestdiff:
                # print(f'bestdiff tie: {bestdiff} | {dataset} {fold} | {f1_metric_error} | {f2_metric_error}')
                tie = True
            if tie:
                f1_wins += 0.5
            elif bestdiff > 0:
                f1_wins += 1
            if tie:
                num_ties += 1
            bestdiff_dataset.append(bestdiff)
        if valid_folds == 0:
            continue
        tierate = num_ties / valid_folds
        bestdiff_dataset = np.mean(bestdiff_dataset)
        f1_winrate = f1_wins / valid_folds
        # print(f'{f1_wins}/{valid_folds} | {dataset}')
        out_dict[dataset] = dict(
            winrate=f1_winrate,
            bestdiff=bestdiff_dataset,
            tierate=tierate,
            wins=f1_wins,
            folds=valid_folds,
        )

    out_df = pd.DataFrame(out_dict).T
    out_df = out_df.sort_values(by=["winrate", "bestdiff", "folds", "tierate"], ascending=[False, False, False, True])
    if verbose:
        print(f"winrate {f1} vs {f2}")
        print(out_df)
    return out_df


def convert_to_dataset_fold(df: pd.DataFrame, column_name="dataset", column_type=str) -> pd.DataFrame:
    """
    Converts "{dataset}_{fold}" dataset column to (dataset, fold) columns.

    Parameters
    ----------
    df: pandas DataFrame of benchmark results
        The DataFrame must have a column named `dataset` in the form "{dataset}_{fold}".
        Example: "adult_5"
    column_name: str, default = "dataset"
        The name of the column to create alongside "fold".
    column_type: type, default = str
        The dtype of the column to create alongside "fold".

    Returns
    -------
    Pandas DataFrame of benchmark results with the dataset column split into (dataset, fold) columns

    """
    df = df.copy(deep=True)
    df[[column_name, "fold"]] = df["dataset"].str.rsplit("_", n=1, expand=True)
    df["fold"] = df["fold"].astype(int)
    df[column_name] = df[column_name].astype(column_type)
    df_columns = [column_name, "fold"] + [c for c in df.columns if c not in [column_name, "dataset", "fold"]]
    return df[df_columns]  # Reorder so dataset and fold are the first two columns.



import json
import logging
import os
import time

import numpy as np
from boto3 import client

logger = logging.getLogger(__name__)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)


def formatted_time():
    current_time = time.localtime()
    formatted_time = time.strftime("%Y%m%dT%H%M%S", current_time)
    return formatted_time


def upload_to_s3(s3_bucket: str, s3_dir: str, local_path: str):
    """Uploads a file or a directory to an S3 bucket.

    Args:
        s3_bucket (str): The name of the S3 bucket to upload the file or directory to.
        s3_dir (str): The S3 path of the directory to upload the file or directory to.
        local_path (str): The local path of the file or directory to upload.

    Returns:
        The S3 path of the uploaded file or directory.
    """
    import boto3

    logging.info("Saving to S3 Bucket %s...", s3_bucket)
    s3 = boto3.client("s3")

    if os.path.isfile(local_path):
        file_name = os.path.basename(local_path)
        s3_path = os.path.join(s3_dir, file_name)
        s3.upload_file(local_path, s3_bucket, s3_path)
        logging.info(f"File {local_path} has been saved to s3://{s3_bucket}/{s3_path}")
        return f"s3://{s3_bucket}/{s3_path}"
    elif os.path.isdir(local_path):
        if len(os.listdir(local_path)) == 0:
            logger.warning(f"No files under {local_path}.")
            return

        for root, dirs, files in os.walk(local_path):
            for filename in files:
                file_local_path = os.path.join(root, filename)

                relative_path = os.path.relpath(file_local_path, local_path)
                s3_path = os.path.join(s3_dir, relative_path)

                s3.upload_file(file_local_path, s3_bucket, s3_path)

        logging.info("Files under %s have been saved to s3://%s/%s", local_path, s3_bucket, s3_dir)
        return f"s3://{s3_bucket}/{s3_dir}"


def download_file_from_s3(s3_path: str, local_path: str = "/tmp") -> str:
    """Downloads a file from an S3 bucket.

    Args:
        s3_path (str): The S3 path of the file.
        local_path (str): The local path where the file will be downloaded.

    Returns:
        str: The local path of the downloaded file.
    """
    import boto3

    logging.info(f"Downloading file from {s3_path} to {local_path}.")
    s3 = boto3.client("s3")

    bucket = s3_path.strip("s3://").split("/")[0]
    s3_path = s3_path[len(f"s3://{bucket}/") :]

    local_file_path = os.path.join(local_path, s3_path.split("/")[-1])
    s3.download_file(bucket, s3_path, local_file_path)

    return local_file_path


def download_dir_from_s3(s3_path: str, local_path: str) -> str:
    """Downloads a directory from an S3 bucket.

    Args:
        s3_path (str): The S3 path of the directory.
        local_path (str): The local path where the directory will be downloaded.

    Returns:
        str: The local path of the downloaded directory.
    """
    import boto3

    logging.info(f"Downloading dir from {s3_path} to {local_path}.")
    s3 = boto3.client("s3")

    bucket = s3_path.strip("s3://").split("/")[0]
    s3_path = s3_path[len(f"s3://{bucket}/") :]

    response = s3.list_objects(Bucket=bucket, Prefix=s3_path)

    for content in response.get("Contents", []):
        s3_obj_path = content["Key"]
        relative_path = os.path.relpath(s3_obj_path, s3_path)
        local_obj_path = os.path.join(local_path, relative_path)

        os.makedirs(os.path.dirname(local_obj_path), exist_ok=True)
        s3.download_file(bucket, s3_obj_path, local_obj_path)

    return local_path



from typing import List, Union

import openml
import pandas as pd

from autogluon.common.savers import save_pd


def generate_metadata(study: Union[int, List[int]]) -> pd.DataFrame:
    """
    Generates the OpenML metadata file for every task in a study.
    """
    tasks = openml.tasks.list_tasks(output_format="dataframe")
    if not isinstance(study, list):
        study = [study]
    suite_tasks = []
    for s in study:
        suite = openml.study.get_suite(s)
        suite_tasks += suite.tasks
    suite_tasks = list(set(suite_tasks))
    suite_tasks.sort()
    task_metadata = tasks[tasks["tid"].isin(set(suite_tasks))]
    return task_metadata


def generate_and_save_metadata(path: str, study: Union[int, List[int]]):
    task_metadata = generate_metadata(study=study)
    save_pd.save(path=path, df=task_metadata)


if __name__ == "__main__":
    # Generate the AutoMLBenchmark study metadata of 104 datasets
    path = "data/metadata/task_metadata.csv"
    generate_and_save_metadata(path=path, study=[269, 271])



#!/usr/bin/env python3

import logging
import os

from aws_cdk import App, Environment, Tags

from autogluon.bench.cloud.aws.batch_stack.stack import BatchJobStack, StaticResourceStack

logger = logging.getLogger(__name__)


def get_mandatory_env(name):
    """
    Reads the env variable, raises an exception if missing.
    """
    if name not in os.environ:
        raise Exception("Missing os enviroment variable '%s'" % name)
    return os.environ.get(name)


cdk_default_account = get_mandatory_env("CDK_DEPLOY_ACCOUNT")
cdk_default_region = get_mandatory_env("CDK_DEPLOY_REGION")
logger.info("Deploying the stack to %s %s", cdk_default_account, cdk_default_region)

app = App()
prefix = app.node.try_get_context("STACK_NAME_PREFIX")
tag = app.node.try_get_context("STACK_NAME_TAG")
static_resource_stack_name = app.node.try_get_context("STATIC_RESOURCE_STACK_NAME")
batch_stack_name = app.node.try_get_context("BATCH_STACK_NAME")

env = Environment(account=cdk_default_account, region=cdk_default_region)
static_resource_stack = StaticResourceStack(app, static_resource_stack_name, env=env)
batch_stack = BatchJobStack(app, batch_stack_name, env=env, static_stack=static_resource_stack)
batch_stack.add_dependency(static_resource_stack)

# Stack-level tag which expects to be propogated to resources
# Looks like Batch (compute environment, job definition, job queue)
# is not supported currently
# https://aws.amazon.com/premiumsupport/knowledge-center/cloudformation-propagate-stack-level-tag/
Tags.of(app).add(key=prefix, value=tag)
app.synth()



import argparse
import csv
import importlib
import json
import logging
import os
import random
import time
from datetime import datetime, timezone
from typing import Optional, Union

import numpy as np
from sklearn.model_selection import train_test_split

from autogluon.bench.datasets.dataset_registry import multimodal_dataset_registry
from autogluon.core.metrics import make_scorer
from autogluon.multimodal import MultiModalPredictor
from autogluon.multimodal import __version__ as ag_version
from autogluon.multimodal.constants import IMAGE_SIMILARITY, IMAGE_TEXT_SIMILARITY, OBJECT_DETECTION, TEXT_SIMILARITY

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def _flatten_dict(data):
    flattened = {}
    for key, value in data.items():
        if isinstance(value, dict):
            flattened.update(_flatten_dict(value))
        else:
            flattened[key] = value
    return flattened


def set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)


def get_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--dataset_name",
        type=str,
        help="Dataset that has been registered with multimodal_dataset_registry.",
    )
    parser.add_argument("--framework", type=str, help="Framework (and) branch/version.")
    parser.add_argument("--benchmark_dir", type=str, help="Directory to save benchmarking run.")
    parser.add_argument("--metrics_dir", type=str, help="Directory to save benchmarking metrics.")
    parser.add_argument("--constraint", type=str, default=None, help="AWS resources constraint setting.")
    parser.add_argument("--params", type=str, default=None, help="AWS resources constraint setting.")
    parser.add_argument(
        "--custom_dataloader", type=str, default=None, help="Custom dataloader to use in the benchmark."
    )
    parser.add_argument("--custom_metrics", type=str, default=None, help="Custom metrics to use in the benchmark.")
    parser.add_argument("--time_limit", type=int, default=None, help="Time limit used to fit the predictor.")

    args = parser.parse_args()
    return args


def load_dataset(dataset_name: str, custom_dataloader: dict = None):  # dataset name
    """Loads and preprocesses a dataset.

    Args:
        dataset_name (str): The name of the dataset to load.
        custom_dataloader (dict): A dictionary containing information about a custom dataloader to use. Defaults to None.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: A tuple containing the training and test datasets.
    """
    splits = ["train", "val", "test"]
    data = {}
    if dataset_name in multimodal_dataset_registry.list_keys():
        logger.info(f"Loading dataset {dataset_name} from multimodal_dataset_registry")
        for split in splits:
            data[split] = multimodal_dataset_registry.create(dataset_name, split)
    elif custom_dataloader is not None:
        logger.info(f"Loading dataset {dataset_name} from custom dataloader {custom_dataloader}.")
        custom_dataloader_file = custom_dataloader.pop("dataloader_file")
        class_name = custom_dataloader.pop("class_name")
        spec = importlib.util.spec_from_file_location(class_name, custom_dataloader_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        custom_class = getattr(module, class_name)
        for split in splits:
            data[split] = custom_class(dataset_name=dataset_name, split=split, **custom_dataloader)
    else:
        raise ModuleNotFoundError(f"Dataset Loader for dataset {dataset_name} is not available.")

    return data.values()


def load_custom_metrics(custom_metrics: dict):
    """Loads a custom metrics and convert it to AutoGluon Scorer.

    Args:
        custom_metrics (dict): A dictionary containing information about a custom metrics to use. Defaults to None.

    Returns:
        scorer (Scorer)
            scorer: An AutoGluon Scorer object to pass to MultimodalPredictor.
    """

    try:
        custom_metrics_path = custom_metrics.pop("metrics_path")
        func_name = custom_metrics.pop("function_name")
        spec = importlib.util.spec_from_file_location(func_name, custom_metrics_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        score_func = getattr(module, func_name)

        scorer = make_scorer(
            name=func_name,
            score_func=score_func,
            **custom_metrics,  # https://auto.gluon.ai/stable/tutorials/tabular/advanced/tabular-custom-metric.html
        )
    except:
        raise ModuleNotFoundError(f"Unable to load custom metrics function {func_name} from {custom_metrics_path}.")

    return scorer


def save_metrics(metrics_path: str, metrics: dict):
    """Saves evaluation metrics to a JSON file.

    Args:
        metrics_path (str): The path to the directory where the metrics should be saved.
        metrics: The evaluation metrics to save.

    Returns:
        None
    """
    if metrics is None:
        logger.warning("No metrics were created.")
        return

    if not os.path.exists(metrics_path):
        os.makedirs(metrics_path)
    file = os.path.join(metrics_path, "results.csv")
    flat_metrics = _flatten_dict(metrics)
    field_names = flat_metrics.keys()

    with open(file, "w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=field_names)
        writer.writeheader()
        writer.writerow(flat_metrics)
    logger.info("Metrics saved to %s.", file)
    f.close()


def run(
    dataset_name: Union[str, dict],
    framework: str,
    benchmark_dir: str,
    metrics_dir: str,
    constraint: Optional[str] = None,
    params: Optional[dict] = None,
    custom_dataloader: Optional[dict] = None,
    custom_metrics: Optional[dict] = None,
    time_limit: Optional[int] = None,
):
    """Runs the AutoGluon multimodal benchmark on a given dataset.

    Args:
        dataset_name (Union[str, dict]): Dataset that has been registered with multimodal_dataset_registry.

                            To get a list of datasets:

                            from autogluon.bench.datasets.dataset_registry import multimodal_dataset_registry
                            multimodal_dataset_registry.list_keys()

        benchmark_dir (str): The path to the directory where benchmarking artifacts should be saved.
        constraint (str): The resource constraint used by benchmarking during AWS mode, default: None.
        params (dict): The multimodal params, default: {}.
        custom_dataloader (dict): A dictionary containing information about a custom dataloader to use. Defaults to None.
                                To define a custom dataloader in the config file:

                                custom_dataloader:
                                    dataloader_file: path_to/dataloader.py   # relative path to WORKDIR
                                    class_name: DataLoaderClass
                                    dataset_config_file: path_to/dataset_config.yaml
                                    **kwargs (of DataLoaderClass)
        custom_metrics (dict): A dictionary containing information about a custom metrics to use. Defaults to None.
                                To define a custom metrics in the config file:

                                custom_metrics:
                                    metrics_path: path_to/metrics.py   # relative path to WORKDIR
                                    function_name: custom_metrics_function
                                    **kwargs (of autogluon.core.metrics.make_scorer)
    Returns:
        None
    """
    seed = params.get("seed", 42)
    set_seed(seed)

    train_data, val_data, test_data = load_dataset(dataset_name=dataset_name, custom_dataloader=custom_dataloader)
    if test_data.data is None:
        logger.warning("No test data found, splitting test data from train data")
        train_set, test_set = train_test_split(train_data.data, test_size=0.2, random_state=seed)
        train_data.data = train_set
        test_data.data = test_set

    try:
        label_column = train_data.label_columns[0]
    except (AttributeError, IndexError):  # Object Detection does not have label columns
        label_column = None
    if params is None:
        params = {}
    predictor_args = {
        "label": label_column,
        "problem_type": train_data.problem_type,
        "presets": params.pop("presets", None),  # backward compatible with <= v0.4.4
        "path": os.path.join(benchmark_dir, "models"),
    }
    predictor_args.update(params.pop("predictor_args", {}))

    if val_data is not None:
        predictor_args["eval_metric"] = val_data.metric

    if train_data.problem_type == IMAGE_SIMILARITY:
        predictor_args["query"] = train_data.image_columns[0]
        predictor_args["response"] = train_data.image_columns[1]
        predictor_args["match_label"] = train_data.match_label
    elif train_data.problem_type == IMAGE_TEXT_SIMILARITY:
        predictor_args["query"] = train_data.text_columns[0]
        predictor_args["response"] = train_data.image_columns[0]
        del predictor_args["label"]
    elif train_data.problem_type == TEXT_SIMILARITY:
        predictor_args["query"] = train_data.text_columns[0]
        predictor_args["response"] = train_data.text_columns[1]
        predictor_args["match_label"] = train_data.match_label
    elif train_data.problem_type == OBJECT_DETECTION:
        predictor_args["sample_data_path"] = train_data.data

    metrics_func = None
    if custom_metrics is not None and custom_metrics["function_name"] == train_data.metric:
        metrics_func = load_custom_metrics(custom_metrics=custom_metrics)

    predictor = MultiModalPredictor(**predictor_args)

    if time_limit is not None:
        params["time_limit"] = time_limit
        logger.warning(
            f'params["time_limit"] is being overriden by time_limit specified in constraints.yaml. params["time_limit"] = {time_limit}'
        )

    fit_args = {"train_data": train_data.data, "tuning_data": val_data.data}
    fit_args.update(params.pop("fit_args", {}))
    fit_args.update(**params)  # backward compatible with <= v0.4.4

    utc_time = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")
    start_time = time.time()
    predictor.fit(**fit_args)
    end_time = time.time()
    training_duration = round(end_time - start_time, 1)

    if isinstance(test_data.data, dict):  # multiple test datasets
        test_data_dict = test_data.data

    else:
        test_data_dict = {dataset_name: test_data}

    for dataset_name, test_data in test_data_dict.items():
        evaluate_args = {
            "data": test_data.data,
            "label": label_column,
            "metrics": test_data.metric if metrics_func is None else metrics_func,
        }

        if test_data.problem_type == IMAGE_TEXT_SIMILARITY:
            evaluate_args["query_data"] = test_data.data[test_data.text_columns[0]].unique().tolist()
            evaluate_args["response_data"] = test_data.data[test_data.image_columns[0]].unique().tolist()
            evaluate_args["cutoffs"] = [1, 5, 10]

        start_time = time.time()
        scores = predictor.evaluate(**evaluate_args)
        end_time = time.time()
        predict_duration = round(end_time - start_time, 1)

        if "#" in framework:
            framework, version = framework.split("#")
        else:
            framework, version = framework, ag_version

        metric_name = test_data.metric if metrics_func is None else metrics_func.name
        primary_metric = metric_name[0] if isinstance(metric_name, list) else metric_name
        if predictor_args.get("use_ensemble"):
            result = scores["ensemble"]
        else:
            result = scores[primary_metric]

        if hasattr(train_data, "id"):
            id = f"id/{train_data.id}"
        else:
            id = "id/0"  # dummy id to make it align with amlb benchmark output
        metrics = {
            "id": id,
            "task": dataset_name,
            "framework": framework,
            "constraint": constraint,
            "version": version,
            "fold": 0,
            "type": predictor.problem_type,
            "metric": primary_metric,
            "result": result,
            "utc": utc_time,
            "training_duration": training_duration,
            "predict_duration": predict_duration,
            "scores": scores,
        }
        subdir = f"{framework}.{dataset_name}.{constraint}.local"
        save_metrics(os.path.join(metrics_dir, subdir, "scores"), metrics)


if __name__ == "__main__":
    args = get_args()
    if args.params is not None:
        args.params = json.loads(args.params)
    if args.custom_dataloader is not None:
        args.custom_dataloader = json.loads(args.custom_dataloader)
    if args.custom_metrics is not None:
        args.custom_metrics = json.loads(args.custom_metrics)

    run(
        dataset_name=args.dataset_name,
        framework=args.framework,
        benchmark_dir=args.benchmark_dir,
        metrics_dir=args.metrics_dir,
        constraint=args.constraint,
        params=args.params,
        custom_dataloader=args.custom_dataloader,
        custom_metrics=args.custom_metrics,
        time_limit=args.time_limit,
    )



import typer
import yaml

app = typer.Typer()


@app.command()
def generate_cloud_config(
    module: str = typer.Option(..., help="The module to generate config for: 'multimodal', 'tabular' or 'timeseries'"),
    root_dir: str = typer.Option("ag_bench_runs", help="Root directory (optional, default = 'ag_bench_runs')"),
    benchmark_name: str = typer.Option("ag_bench", help="Benchmark name (optional, default = 'ag_bench')"),
    cdk_deploy_account: str = typer.Option(..., help="CDK deploy account"),
    cdk_deploy_region: str = typer.Option(..., help="CDK deploy region"),
    prefix: str = typer.Option(
        "ag-bench", help="Prefix used to identify infra resources (optional, default = 'ag-bench')"
    ),
    metrics_bucket: str = typer.Option(..., help="Name of the metrics bucket (required)"),
    data_bucket: str = typer.Option(None, help="S3 bucket to download private datasets (optional)"),
    max_machine_num: int = typer.Option(
        None, help="Maximum number of machines for AWS Batch (optional, default = 20)"
    ),
    block_device_volume: int = typer.Option(
        None, help="Block device volume for EC2 instances (optional, default = 100)"
    ),
    reserved_memory_size: int = typer.Option(
        None, help="Reserved memory size for Docker container (optional, default = 15000)"
    ),
    instance: str = typer.Option(None, help="EC2 Instance type (optional, default = 'g4dn.2xlarge')"),
    time_limit: str = typer.Option(
        None,
        help="AWS Batch job time out",
    ),
    vpc_name: str = typer.Option(None, help="Existing VPC name (optional, default: create a new VPC)"),
    git_uri_branch: str = typer.Option("", help="AMLB git_uri#branch"),
    dataset_names: str = typer.Option(
        "",
        help="AutoGluon MultiModal dataset names for '--module multimodal' (comma-separated, to get a list of dataset names, run `from autogluon.bench.datasets.dataset_registry import multimodal_dataset_registry\n multimodal_dataset_registry.list_keys()`)",
    ),
    custom_resource_dir: str = typer.Option(
        None,
        help="Path to custom resources definitions for '--module multimodal'",
    ),
    custom_dataloader: str = typer.Option(
        None,
        help="Custom dataloader for '--module multimodal', in the format '\"dataloader_file:value1;class_name:value2;dataset_config_file:value3\"'",
    ),
    framework: str = typer.Option(
        "AutoGluon:stable",
        help="Framework name",
    ),
    constraint: str = typer.Option("test", help="Resource constraint for '--module multimodal'"),
    fewshot: bool = typer.Option(False, help="Flag to enable/disable fewshot learning for '--module multimodal'"),
    shot: int = typer.Option(0, help="Number of shots if fewshot learning is enabled '--module multimodal'"),
    lang: str = typer.Option("en", help="Language for datasets '--module multimodal text datasets'"),
    seed: int = typer.Option(0, help="Seed value '--module multimodal'"),
    amlb_constraint: str = typer.Option(
        "",
        help="AMLB Constraints for tabular or timeseries, in the format 'test,1h4c,...'. Refer to https://github.com/openml/automlbenchmark/blob/master/resources/constraints.yaml for details.",
    ),
    amlb_benchmark: str = typer.Option(
        "",
        help="AMLB Benchmarks for tabular or timeseries, in the format 'test,small,...'. Refer to https://github.com/openml/automlbenchmark/tree/master/resources/benchmarks for options.",
    ),
    amlb_task: str = typer.Option(
        None,
        help="AMLB Tasks for tabular or timeseries (in the format '\"benchmark1:task1,task2;benchmark2:task3,task4,task5;...\"')",
    ),
    amlb_fold_to_run: str = typer.Option(
        None,
        help="AMLB fold for tabular or timeseries (in the format '\"benchmark1:task1:fold1/fold2,task2:fold3/fold4;benchmark2:task3:fold1/fold2;...\"')",
    ),
    amlb_user_dir: str = typer.Option(
        None,
        help="Custom config directory.",
    ),
    custom_metrics: bool = typer.Option(False, help="Flag to enable/disable custom_metrics for '--module multimodal'"),
    metrics_path: str = typer.Option(None, help="Resource constraint for '--module multimodal custom metrics'"),
    function_name: str = typer.Option(None, help="Resource constraint for '--module multimodal custom metrics'"),
    optimum: int = typer.Option(0, help="Resource constraint for '--module multimodal custom metrics'"),
    greater_is_better: bool = typer.Option(
        False,
        help="Flag to enable/disable resource constraint for '--module multimodal'",
    ),
):
    config = {
        "module": module,
        "mode": "aws",
        "benchmark_name": benchmark_name,
        "root_dir": root_dir,
        "framework": framework,
    }
    cdk_context = {
        "CDK_DEPLOY_ACCOUNT": cdk_deploy_account,
        "CDK_DEPLOY_REGION": cdk_deploy_region,
        "PREFIX": prefix,
        "METRICS_BUCKET": metrics_bucket,
    }
    if data_bucket:
        cdk_context["DATA_BUCKET"] = data_bucket
    if vpc_name:
        cdk_context["VPC_NAME"] = vpc_name
    if max_machine_num:
        cdk_context["MAX_MACHINE_NUM"] = max_machine_num
    if block_device_volume:
        cdk_context["BLOCK_DEVICE_VOLUME"] = block_device_volume
    if reserved_memory_size:
        cdk_context["RESERVED_MEMORY_SIZE"] = reserved_memory_size
    if instance:
        cdk_context["INSTANCE"] = instance
    if time_limit:
        cdk_context["TIME_LIMIT"] = time_limit

    config["cdk_context"] = cdk_context

    if module == "multimodal":
        dataset_names = dataset_names.split(",") if dataset_names else None
        module_configs = {
            "constraint": constraint,
            "dataset_name": dataset_names,
        }
        if custom_resource_dir:
            module_configs["custom_resource_dir"] = custom_resource_dir

        if custom_dataloader:
            custom_dataloader_dict = {}
            for item in custom_dataloader.split(";"):
                k, v = item.split(":")
                custom_dataloader_dict[k] = v
            module_configs["custom_dataloader"] = custom_dataloader_dict
            if fewshot:
                module_configs["custom_dataloader"]["fewshot"] = fewshot
                module_configs["custom_dataloader"]["lang"] = lang
                module_configs["custom_dataloader"]["shot"] = shot
                module_configs["custom_dataloader"]["seed"] = seed

        if custom_metrics:
            module_configs["custom_metrics"] = {}
            module_configs["custom_metrics"]["metrics_path"] = metrics_path
            module_configs["custom_metrics"]["function_name"] = function_name
            module_configs["custom_metrics"]["optimum"] = optimum
            module_configs["custom_metrics"]["greater_is_better"] = greater_is_better

        config.update(module_configs)

    elif module in ["tabular", "timeseries"]:
        module_configs = {
            "git_uri#branch": git_uri_branch,
            "amlb_constraint": amlb_constraint,
        }

        if amlb_benchmark:
            amlb_benchmark = amlb_benchmark.split(",")
            module_configs["amlb_benchmark"] = amlb_benchmark

        if amlb_task:
            amlb_task_dict = {}
            for item in amlb_task.split(";"):
                if ":" in item:
                    benchmark, tasks = item.split(":")
                    task_list = tasks.split(",")
                    amlb_task_dict[benchmark] = task_list
            module_configs["amlb_task"] = amlb_task_dict

        if amlb_fold_to_run:
            fold_to_run_dict = {}
            for benchmark_item in amlb_fold_to_run.split(";"):
                benchmark, task_item = benchmark_item.split(":", 1)
                fold_to_run_dict[benchmark] = {}
                for task_folds in task_item.split(","):
                    task, folds = task_folds.split(":")
                    folds = [int(v) for v in folds.split("/")]
                    fold_to_run_dict[benchmark][task] = folds
            module_configs["fold_to_run"] = fold_to_run_dict

        if amlb_user_dir:
            module_configs["amlb_user_dir"] = amlb_user_dir

        config.update(module_configs)
    else:
        typer.echo("Invalid module. Please choose 'multimodal', 'tabular' or 'timeseries'.")
        return

    output_file = f"{module}_cloud_configs.yaml"
    with open(output_file, "w") as f:
        yaml.dump(config, f)

    typer.echo(f"Config file '{output_file}' generated successfully.")
